"""Downlad stats upload to vol_txt in XNAT."""
import logging
import tempfile
import csv


logger = logging.getLogger('garjus.automations.xnat_ma3stats2voltxt')


def process_project(
    xnat,
    project,
    assessors,
):
    """Process project."""
    results = []

    for i, assr in assessors.iterrows():
        if 'VOL_TXT' in assr['RESOURCES'].split(','):
            logger.debug(f'VOL_TXT found in RESOURCES:{assr.ASSR}')
            continue

        if 'STATS' not in assr['RESOURCES'].split(','):
            logger.debug(f'STATS not found in RESOURCES not:{assr.ASSR}')
            continue

        # Download stats, transform, upload
        with tempfile.TemporaryDirectory() as tmpdir:
            vol_file = f'{tmpdir}/target_processed_label_volumes.txt'
            stats_file = f'{tmpdir}/stats.csv'

            # Extract stats file
            logger.debug('download stats')
            xnat.select_assessor_resource(
                assr['PROJECT'],
                assr['SUBJECT'],
                assr['SESSION'],
                assr['ASSR'],
                'STATS'
            ).file('stats.csv').get(stats_file)

            # Transform it
            stats2voltxt(stats_file, vol_file)

            # Upload it
            logger.debug('upload voltxt')
            xnat.select_assessor_resource(
                assr['PROJECT'],
                assr['SUBJECT'],
                assr['SESSION'],
                assr['ASSR'],
                'VOL_TXT'
            ).file('target_processed_label_volumes.txt').put(vol_file)

        results.append({
            'result': 'COMPLETE',
            'category': 'ma3stats2voltxt',
            'description': assr['ASSR'],
            'subject': assr['SUBJECT'],
            'session': assr['SESSION'],
        })

    return results


def stats2voltxt(stats, voltxt):
    with open(stats, newline='') as f:
        reader = csv.reader(f)
        keys = next(reader)
        values = next(reader)

    # Remove first two items that are not found in MA v2
    keys = keys[2:]
    values = values[2:]

    with open(voltxt, 'w') as f:
        f.write('Name,Name,Volume\n')
        for i, k in enumerate(keys):
            f.writelines(f'{k},{k},{values[i]}\n')



import logging

import pandas as pd
import plotly.graph_objs as go
import plotly.subplots
from dash import dcc, html, dash_table as dt
from dash import Input, Output
import dash_bootstrap_components as dbc

from ..app import app
from .. import utils
from ..shared import GWIDTH, STATUS2RGB
from . import data


logger = logging.getLogger('dashboard.issues')


def _get_graph_content(df):
    PIVOTS = ['PROJECT', 'CATEGORY']
    STATUSES = ['FAIL', 'COMPLETE', 'PASS', 'UNKNOWN']
    status2rgb = {k: STATUS2RGB[k] for k in STATUSES}
    tabs_content = []

    # index we are pivoting on to count statuses
    for i, pindex in enumerate(PIVOTS):
        # Make a 1x1 figure
        fig = plotly.subplots.make_subplots(rows=1, cols=1)
        fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

        # Draw bar for each status, these will be displayed in order
        dfp = pd.pivot_table(
            df,
            index=pindex,
            values='LABEL',
            columns=['STATUS'],
            aggfunc='count',
            fill_value=0)

        for status, color in status2rgb.items():
            ydata = sorted(dfp.index)
            if status not in dfp:
                xdata = [0] * len(dfp.index)
            else:
                xdata = dfp[status]

            fig.append_trace(go.Bar(
                x=ydata,
                y=xdata,
                name='{} ({})'.format(status, sum(xdata)),
                marker=dict(color=color),
                opacity=0.9), 1, 1)

        # Customize figure
        fig['layout'].update(barmode='stack', showlegend=True, width=GWIDTH)

        # Build the graph in a tab
        label = 'By {}'.format(pindex)
        graph = dcc.Graph(figure=fig)
        tab = dbc.Tab(label=label, children=[graph])

        # Append the tab
        tabs_content.append(tab)

    return dbc.Spinner(
        id="loading-issues",
        children=dbc.Tabs(id='tabs-issues', children=tabs_content))


def get_content():
    columns = utils.make_columns([
        'DATETIME',
        'CATEGORY',
        'PROJECT', 
        'SUBJECT',
        'SESSION',
        'EVENT',
        'DESCRIPTION',
    ])
    # removes: SCAN, FIELD, STATUS, ID

    # Format columns
    for i, c in enumerate(columns):
        if c['name'] in ['PROJECT']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    content = [
        dbc.Row(html.Div(id='container-issues-graph', children=[])),
        dbc.Row([
            dbc.Col(
                dbc.Button(
                    'Refresh Data',
                    id='button-issues-refresh',
                    outline=True,
                    color='primary',
                    size='sm',
                ),
            ),
            dbc.Col(
                dbc.Switch(
                    id='switch-issues-graph',
                    label='Graph',
                    value=False,
                ),
                align='center',
            ),
        ]),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-issues-project',
                    multi=True,
                    placeholder='Select Projects',
                ),
                width=4,
            ),
        ),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-issues-category',
                    multi=True,
                    placeholder='Select Categories',
                ),
                width=4,
            ),
        ),
        dbc.Spinner(id="loading-issues-table", children=[
            dbc.Label('Loading...', id='label-issues-rowcount1'),
        ]),
        dt.DataTable(
            columns=columns,
            data=[],
            filter_action='native',
            page_action='none',
            sort_action='native',
            id='datatable-issues',
            style_table={
                'overflowY': 'scroll',
                'overflowX': 'scroll',
            },
            style_cell={
                'textAlign': 'left',
                'padding': '5px 5px 0px 5px',
                'width': '30px',
                'overflow': 'hidden',
                'textOverflow': 'ellipsis',
                'height': 'auto',
                'minWidth': '40',
                'maxWidth': '60'},
            style_data_conditional=[
                {'if': {'column_id': 'CATEGORY'}, 'textAlign': 'center'},
            ],
            style_header={
                'fontWeight': 'bold',
                'padding': '5px 15px 0px 10px',
            },
            style_cell_conditional=[
                {'if': {'column_id': 'DATETIME'}, 'textAlign': 'center'},
                {'if': {'column_id': 'SUBJECT'}, 'textAlign': 'center'},
                {'if': {'column_id': 'SESSION'}, 'textAlign': 'center'},
                {'if': {'column_id': 'EVENT'}, 'textAlign': 'center'},
            ],
            css=[dict(selector="p", rule="margin: 0; text-align: center")],
            export_format='xlsx',
            export_headers='names',
            export_columns='visible',
        ),
        dbc.Label('Get ready...', id='label-issues-rowcount2'),
    ]

    return content


def load_options(df):
    options = {}

    for k in ['CATEGORY', 'PROJECT']:
        # Get a unique list of strings with blanks removed
        koptions = df[k].unique()
        koptions = [str(x) for x in koptions]
        koptions = [x for x in koptions if x]
        options[k] = sorted(koptions)

    return options


def load_issues(refresh=False):
    return data.load_data(refresh=refresh)


def filter_data(df, selected_project, selected_category):
    return data.filter_data(
        df, selected_project, selected_category)


# Issues callback
@app.callback(
    [
     Output('dropdown-issues-category', 'options'),
     Output('dropdown-issues-project', 'options'),
     Output('datatable-issues', 'data'),
     Output('container-issues-graph', 'children'),
     Output('label-issues-rowcount1', 'children'),
     Output('label-issues-rowcount2', 'children'),
    ],
    [
     Input('dropdown-issues-category', 'value'),
     Input('dropdown-issues-project', 'value'),
     Input('switch-issues-graph', 'value'),
     Input('button-issues-refresh', 'n_clicks'),
    ],
)
def update_issues(
    selected_category,
    selected_project,
    selected_graph,
    n_clicks
):
    refresh = False
    graph_content = []

    logger.debug('update_issues')

    # Load issues data
    if utils.was_triggered('button-issues-refresh'):
        # Refresh data if refresh button clicked
        logger.debug(f'issues refresh:clicks={n_clicks}')
        refresh = True

    logger.debug(f'loading issues data:refresh={refresh}')
    df = load_issues(refresh=refresh)

    # Update lists of possible options for dropdowns (could have changed)
    # make these lists before we filter what to display
    options = load_options(df)
    projects = utils.make_options(options['PROJECT'])
    categories = utils.make_options(options['CATEGORY'])

    # Filter data based on dropdown values
    df = filter_data(
        df,
        selected_project,
        selected_category,
    )

    logger.debug(f'selected_graph:{selected_graph}')
    if selected_graph:
        logger.debug('getting issues graph')
        graph_content = _get_graph_content(df)

    # Get the table data
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
        if r['PROJECT'] and 'PROJECTLINK' in r:
            _proj = r['PROJECT']
            _link = r['PROJECTLINK']
            r['PROJECT'] = f'[{_proj}]({_link})'

    # Return table, figure, dropdown options
    logger.debug('update_issues:returning data')

    # Count how many rows are in the table
    if len(records) > 1:
        rowcount = '{} rows'.format(len(records))
    else:
        rowcount = ''

    return [categories, projects, records, graph_content, rowcount, rowcount]



import click
import pprint
import logging

from .garjus import Garjus

logging.basicConfig(
    format='%(asctime)s - %(levelname)s:%(name)s:%(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.option('--quiet/--no-quiet', default=False)
def cli(debug, quiet):
    if debug:
        click.echo('garjus! debug')
        logging.getLogger().setLevel(logging.DEBUG)

    if quiet:
        logging.getLogger().setLevel(logging.ERROR)
        logging.getLogger('werkzeug').setLevel(logging.ERROR)
        logging.getLogger('dash').setLevel(logging.ERROR)


@cli.command('copysess')
@click.argument('src', required=True)
@click.argument('dst', required=True)
def copy_session(src, dst):
    click.echo('garjus! copy session')
    Garjus().copy_sess(src, dst)


@cli.command('setsesstype')
@click.argument('src', required=True)
@click.argument('sesstype', required=True)
def set_sesstype(src, sesstype):
    click.echo('garjus! set session type')
    Garjus().set_session_type(src, sesstype)


@cli.command('setsite')
@click.argument('src', required=True)
@click.argument('site', required=True)
def set_site(src, site):
    click.echo('garjus! set session site')
    Garjus().set_session_site(src, site)


@cli.command('issues')
@click.option('--project', '-p', 'project')
@click.pass_context
def issues(ctx, project):
    click.echo('garjus! issues')
    g = Garjus()
    pprint.pprint(g.issues(project))


@cli.command('subjects')
@click.option('--project', '-p', 'project')
@click.pass_context
def subjects(ctx, project):
    click.echo('garjus! subjects')
    g = Garjus()
    import pandas as pd
    pd.set_option('display.max_rows', None)
    pprint.pprint(g.subjects(project))


@cli.command('activity')
@click.option('--project', '-p', 'project')
def activity(project):
    click.echo('garjus! activity')
    g = Garjus()
    pprint.pprint(g.activity(project))


@cli.command('analyses')
@click.option('--project', '-p', 'project')
def analyses(project):
    click.echo('garjus! analyses')
    g = Garjus()
    pprint.pprint(g.analyses(project))


@cli.command('getinputs')
@click.argument('analysis_id', required=True)
@click.argument('download_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
def getinputs(project, analysis_id, download_dir):
    click.echo('garjus! getinputs')
    g = Garjus()
    g.get_analysis_inputs(project, analysis_id, download_dir)


@cli.command('run')
@click.argument('analysis_id', required=True)
@click.argument('output_zip', required=False)
@click.option('--project', '-p', 'project', required=True)
def run(project, analysis_id, output_zip):
    click.echo('garjus! run')
    g = Garjus()
    g.run_analysis(project, analysis_id, output_zip)


@cli.command('tasks')
def tasks():
    click.echo('garjus! tasks')
    g = Garjus()
    pprint.pprint(g.tasks())


@cli.command('q2d')
def q2d():
    click.echo('garjus! q2d')
    Garjus().queue2dax()


@cli.command('d2q')
def d2q():
    click.echo('garjus! d2q')
    Garjus().dax2queue()


@cli.command('update')
@click.argument(
    'choice',
    type=click.Choice([
        'stats',
        'issues',
        'progress',
        'automations',
        'compare',
        'tasks',
        'analyses'
    ]),
    required=False,
    nargs=-1)
@click.option('--project', '-p', 'project', multiple=True)
@click.option('--types', '-t', 'types', multiple=True, required=False)
def update(choice, project, types):
    click.echo('garjus! update')
    g = Garjus()
    g.update(projects=project, choices=choice, types=types)
    click.echo('ALL DONE!')


@cli.command('progress')
@click.option('--project', '-p', 'project')
def progress(project):
    click.echo('garjus! progress')
    if project:
        project = project.split(',')

    g = Garjus()
    pprint.pprint(g.progress(projects=project))


@cli.command('processing')
@click.option('--project', '-p', 'project', required=True)
def processing(project):
    click.echo('garjus! processing')

    g = Garjus()
    pprint.pprint(g.processing_protocols(project))


@cli.command('report')
@click.option('--project', '-p', 'project', required=True)
@click.option('--monthly/--no-monthly', default=True, required=False)
def report(project, monthly):
    click.echo('garjus! report')
    Garjus().report(project, monthly=monthly)


@cli.command('stats')
@click.option('--projects', '-p', 'projects', required=True)
@click.option('--types', '-t', 'proctypes', required=False)
@click.option('--sesstypes', '-s', 'sesstypes', required=False)
@click.option('--analysis', '-a', 'analysis', required=False)
@click.option('--persubject', is_flag=True)
@click.argument('csv', required=True)
def stats(projects, proctypes, sesstypes, csv, persubject, analysis):
    click.echo('garjus! stats')
    Garjus().export_stats(
        projects,
        proctypes,
        sesstypes,
        csv,
        persubject,
        analysis)


@cli.command('compare')
@click.option('--project', '-p', 'project', required=True)
def compare(project):
    click.echo('garjus! compare')
    Garjus().compare(project)


@cli.command('importdicom')
@click.argument('src', required=True)
@click.argument('dst', required=True)
def import_dicom(src, dst):
    click.echo('garjus! import')
    g = Garjus()
    g.import_dicom(src, dst)


@cli.command('pdf')
@click.argument('src', required=True)
@click.option('--project', '-p', 'project', required=True)
def export_pdf(src, project):
    click.echo('garjus! pdf')
    g = Garjus()
    g.pdf(src, project)


@cli.command('image03csv')
@click.option('--project', '-p', 'project', required=True)
@click.option(
    '--start', '-s', 'startdate', type=click.DateTime(formats=['%Y-%m-%d']))
@click.option(
    '--end', '-e', 'enddate', type=click.DateTime(formats=['%Y-%m-%d']))
def image03csv(project, startdate, enddate):
    click.echo('garjus! image03csv')
    g = Garjus()
    g.image03csv(project, startdate, enddate)


@cli.command('retry')
@click.option('--project', '-p', 'project', required=True)
def retry(project):
    click.echo('garjus! retry')
    g = Garjus()
    g.retry(project)


@cli.command('image03download')
@click.argument('image03_csv', required=True)
@click.argument('download_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
def image03download(project, image03_csv, download_dir):
    click.echo('garjus! image03download')
    g = Garjus()
    g.image03download(project, image03_csv, download_dir)


@cli.command('delete')
@click.option('--project', '-p', 'project', required=True)
@click.option('--type', '-t', 'proctype', required=True)
def delete(project, proctype):
    click.echo('garjus! delete')
    g = Garjus()
    g.delete_proctype(project, proctype)


@cli.command('dashboard')
@click.option('--auth', 'auth_file', required=False)
def dashboard(auth_file=None):

    from .dashboard import app
    import webbrowser
    url = 'http://localhost:8050'

    if auth_file:
        import yaml
        import dash_auth

        # Load user passwords to use dash's basic authentication
        with open(auth_file, 'rt') as file:
            _data = yaml.load(file, yaml.SafeLoader)
            dash_auth.BasicAuth(app, _data['VALID_USERNAME_PASSWORD_PAIRS'])

    # Open URL in a new tab, if a browser window is already open.
    webbrowser.open_new_tab(url)

    # start up a dashboard app
    app.run_server(host='0.0.0.0')

    print('dashboard app closed!')


@cli.command('quicktest')
def quicktest():
    print('hi')
    click.echo('garjus!')

    try:
        redcap_project = Garjus._default_redcap()
    except Exception:
        print('could not connect to REDCap')
        return

    try:
        xnat_interface = Garjus._default_xnat()
    except Exception:
        print('could not connect to XNAT')
        return

    try:
        g = Garjus(redcap_project, xnat_interface)
    except Exception:
        print('something went wrong')
        return

    g.scans()
    print('all good!')



"""garjus queue 2 dax queue."""
import shutil
import logging
import os
import json

from dax import cluster
from .processors import load_from_yaml


logger = logging.getLogger('garjus2dax')


# This is a temporary bridge between garjus and dax.

# This must run with access to garjus redcap to read the queue
# and access to dax diskq directory to write slurm/processor_spec files.
# It does not need XNAT access nor access to any individual project REDCaps,
# the only REDCap it needs is garjus/ccmutils.
# It can run under a user that has access to the diskq directory but
# does not need to be the owner nor does it need to be the user that runs dax,
# it must be able set the files to the user group that runs dax.
# All info needed comes from REDCap, does not read any local files, only
# writes. Should not need to access XNAT.
# Read these from REDCap for those where status is QUEUED
# then set status to JOB_RUNNING. (will already be JOB_RUNNING in XNAT as
# set when the assessor was created by garjus update tasks)


JOBDIR = '/tmp'
IMAGEDIR = '/data/mcr/centos7/singularity'
RESDIR = '/nobackup/vuiis_daily_singularity/Spider_Upload_Dir'
RUNGROUP = 'h_vuiis'
HOST = 'https://xnat2.vanderbilt.edu/xnat'
USER = 'daxspider'
TEMPLATE = '/data/mcr/centos7/dax_templates/job_template_v3.txt'


def _write_processor_spec(
    filename,
    yaml_file,
    singularity_imagedir,
    job_template,
    user_inputs=None
):

    # Write a file with the path to the base processor and any overrides.
    # The file is intended to be written to diskq using the assessor
    # label as the filename. These paths allow dax to read the yaml file
    # during upload to read the description, etc. and put in PDF.

    with open(filename, 'w') as f:
        # write processor yaml filename
        f.write(f'{yaml_file}\n')

        # write customizations
        if user_inputs:
            for k, v in user_inputs.items():
                f.write(f'{k}={v}\n')

        # singularity_imagedir
        f.write(f'singularity_imagedir={singularity_imagedir}\n')

        # job_template
        f.write(f'job_template={job_template}\n')

        # extra blank line
        f.write('\n')


def _task2dax(
    xnat,
    assr,
    walltime,
    memreq,
    yaml_file,
    user_inputs,
    inputlist,
    var2val
):
    '''Writes a task to a dax slurm script in the local diskq.'''

    # NOTE: this function does the same work as dax task.build_task()
    # at build_cmds() which calls processor.build_cmds(), at the point where we have
    # var2val and the next step is to get the text...
    # more specifically we are splitting build_cmds() where it calls build_text()

    jobdir = JOBDIR
    imagedir = IMAGEDIR
    resdir = RESDIR
    job_rungroup = RUNGROUP
    xnat_host = HOST
    xnat_user = USER
    job_template = TEMPLATE

    batch_file = f'{resdir}/DISKQ/BATCH/{assr}.slurm'
    outlog = f'{resdir}/DISKQ/OUTLOG/{assr}.txt'
    processor_spec_path = f'{resdir}/DISKQ/processor/{assr}'
    assr_dir = f'{jobdir}/{assr}'
    dstdir = f'{resdir}/{assr}'

    # Check for image dir before we give to dax
    if not os.path.isdir(imagedir):
        raise FileNotFoundError(f'singularity images not found:{imagedir}')

    if not os.path.isdir(resdir):
        raise FileNotFoundError(f'upload directory not found:{resdir}')

    if not os.path.isfile(job_template):
        raise FileNotFoundError(f'job template not found:{job_template}')

    for i in inputlist:
        i['fpath'] = i['fpath'].replace('xnat.vanderbilt', 'xnat2.vanderbilt')

    # Load the processor
    processor = load_from_yaml(
        xnat,
        yaml_file,
        user_inputs=user_inputs,
        singularity_imagedir=imagedir,
        job_template=job_template)

    # Build the command text
    cmds = processor.build_text(
        var2val,
        inputlist,
        assr_dir,
        dstdir,
        xnat_host,
        xnat_user)

    logger.info(f'writing batch file:{batch_file}')
    batch = cluster.PBS(
        batch_file,
        outlog,
        [cmds],
        walltime,
        mem_mb=memreq,
        ppn=1,
        env=None,
        email=None,
        email_options='FAIL',
        rungroup=job_rungroup,
        xnat_host=xnat_host,
        job_template=job_template)

    batch.write()

    # Write processor spec file for version 3
    logger.info(f'writing processor spec file:{processor_spec_path}')

    _write_processor_spec(
        processor_spec_path,
        yaml_file,
        imagedir,
        job_template,
        user_inputs)

    # Set group ownership
    shutil.chown(batch_file, group='h_vuiisadmin')
    shutil.chown(processor_spec_path, group='h_vuiisadmin')


def queue2dax(garjus):

    # Get the current task table from garjus
    tasks = garjus.tasks()

    # Update each task
    for i, t in tasks.iterrows():
        assr = t['ASSESSOR']
        status = t['STATUS']

        if status not in ['JOB_QUEUED', 'QUEUED']:
            logger.debug(f'skipping:{i}:{assr}:{status}')
            continue

        logger.info(f'{i}:{assr}:{status}')

        walltime = t['WALLTIME']
        memreq = t['MEMREQ']
        inputlist = json.loads(t['INPUTLIST'], strict=False)
        var2val = json.loads(t['VAR2VAL'], strict=False)
        yaml_file = t['YAMLFILE']
        user_inputs = t['USERINPUTS']

        try:
            # Locate the yaml file
            if yaml_file == 'CUSTOM':
                # Download it locally
                yaml_file = garjus.save_task_yaml(
                    t['PROJECT'], t['ID'], f'{RESDIR}/DISKQ/processor')
                shutil.chown(yaml_file, group='h_vuiisadmin')
            else:
                # We already have a local copy so point to it
                yaml_file = os.path.join(garjus._yamldir, yaml_file)

            _task2dax(
                garjus.xnat(),
                assr,
                walltime,
                memreq,
                yaml_file,
                user_inputs,
                inputlist,
                var2val)

            garjus.set_task_status(t['PROJECT'], t['ID'], 'JOB_RUNNING')

        except Exception as err:
            logger.error(err)
            import traceback
            traceback.print_exc()




"""dash index page."""
import logging

from dash import html
import dash_bootstrap_components as dbc

from .app import app
from . import qa
from . import activity
from . import issues
from . import queue
from . import stats
from . import analyses
from . import processors
from . import reports
from ..garjus import Garjus


logger = logging.getLogger('garjus.dashboard')


footer_content = [
    html.Hr(),
    html.Div(
        [
            html.A(
                "garjus",
                href='https://github.com/ccmvumc/garjus',
                target="_blank",
            )
        ],
        style={'textAlign': 'center'},
    ),
]

if Garjus().redcap_enabled():
    tabs = dbc.Tabs([
        dbc.Tab(
            label='QA',
            tab_id='tab-qa',
            children=qa.get_content(),
        ),
        dbc.Tab(
            label='Issues',
            tab_id='tab-issues',
            children=issues.get_content(),
        ),
        dbc.Tab(
            label='Queue',
            tab_id='tab-queue',
            children=queue.get_content(),
        ),
        dbc.Tab(
            label='Activity',
            tab_id='tab-activity',
            children=activity.get_content(),
        ),
        dbc.Tab(
            label='Stats',
            tab_id='tab-stats',
            children=stats.get_content(),
        ),
        dbc.Tab(
            label='Processors',
            tab_id='tab-processors',
            children=processors.get_content(),
        ),
        dbc.Tab(
            label='Reports',
            tab_id='tab-reports',
            children=reports.get_content(),
        ),
        dbc.Tab(
            label='Analyses',
            tab_id='tab-analyses',
            children=analyses.get_content(),
        ),
    ])
else:
    tabs = html.Div(qa.get_content())

app.layout = html.Div(
    className='dbc',
    style={'marginLeft': '20px', 'marginRight': '20px'},
    children=[
        html.Div(id='report-content', children=[tabs]),
        html.Div(id='footer-content', children=footer_content)
    ])

# For gunicorn to work correctly
server = app.server

# Allow external css
app.css.config.serve_locally = False

# Set the title to appear on web pages
app.title = 'dashboard'

if __name__ == '__main__':
    app.run_server(host='0.0.0.0')



"""Garjus audit imaging."""
import logging


logger = logging.getLogger('garjus.issues.audit_imaging')


def audit(scan_table, src_labels, dst_labels):
    """find issues with scan archiving."""
    results = []

    # Process each record
    for r in scan_table:
        result = _audit_record(r, src_labels, dst_labels)
        if result:
            results.append(result)

    return results


def _audit_record(record, src_labels, dst_labels):
    src_sess = record['src_session']
    dst_subj = record['dst_subject']
    dst_sess = record['dst_session']
    scandate = record['scandate']
    event = record['event']
    result = {}

    if dst_subj.startswith('v'):
        dst_subj = dst_subj.replace('v', 'V')

    if dst_sess.startswith('v'):
        # Hard code this to avoid duplicates
        dst_sess = dst_sess.replace('v', 'V')

    # Remove PI prefix if present
    if '_' in src_sess:
        logger.debug(f'{src_sess}:removing PI prefix')
        src_sess = src_sess.split('_')[1]

    # Find problems
    if dst_sess in dst_labels:
        # Ignore record if destination session already exists
        logger.debug(f'{dst_sess}:already archived')
        return None
    elif not src_sess:
        msg = f'{dst_subj}:source session not set'
        logger.debug(msg)
        result.update({'category': 'MISSING_VALUE', 'description': msg})
    elif not dst_sess:
        msg = f'{dst_subj}:destination session not set'
        logger.debug(msg)
        result.update({'category': 'MISSING_VALUE', 'description': msg})
    elif not scandate:
        # We are logging but ignoring this scenario which means we would not
        # catch the case where user forgets to enter the session date.
        msg = f'{dst_subj}:{dst_sess}:date not set'
        logger.debug(msg)
        return None
    elif src_sess not in src_labels:
        # Check that session does actually exist in source project
        msg = f'{src_sess}:not in source project'
        logger.debug(msg)
        result.update({'category': 'MISSING_SESSION', 'description': msg})
    elif dst_sess not in dst_labels:
        # Add issue that auto archive needs to run
        msg = f'{src_sess}:auto archive not working'
        logger.debug(msg)
        result.update({'category': 'NEEDS_AUTO', 'description': msg})

    # Add other details
    result.update({
        'subject': dst_subj,
        'session': dst_sess,
        'event': event,
        'date': scandate
    })

    return result



"""dcm2niix scans in XNAT."""
import logging
import subprocess as sb
import tempfile
import os
import pathlib
import glob

from .. import utils_dcm2nii
from .. import utils_xnat


logger = logging.getLogger('garjus.automations.xnat_dcm2niix')


def process_project(
    garjus,
    project,
):
    """xnat dcm2niix."""
    results = []

    logger.debug(f'loading data:{project}')
    df = garjus.scans(projects=[project])

    # Check each scan
    for i, scan in df.iterrows():
        full_path = scan['full_path']

        if scan['QUALITY'] == 'unusable':
            logger.debug(f'skipping unusable:{scan.SESSION}:{scan.SCANID}')
            continue

        if 'NIFTI' in scan['RESOURCES']:
            logger.debug(
                f'NIFTI exists:{project}:{scan.SESSION}:{scan.SCANID}')
            continue

        if 'JSON' in scan['RESOURCES']:
            logger.debug(f'JSON exists:{project}:{scan.SESSION}:{scan.SCANID}')
            continue

        if 'DICOMZIP' in scan['RESOURCES']:
            logger.debug(f'DICOMZIP to NIFTI:{full_path}')
            try:
                _dicomzip2nifti(garjus, full_path)
            except Exception as err:
                logger.error(err)
                continue
        elif 'DICOM' in scan['RESOURCES']:
            logger.debug(f'No DICOMZIP found, using DICOM:{full_path}')
            try:
                _dicomdir2nifti(garjus, full_path)
            except Exception as err:
                logger.error(err)
                continue
        else:
            logger.debug(f'no DICOMZIP or DICOM:{full_path}')
            continue

        results.append({
            'result': 'COMPLETE',
            'description': 'dcm2niix',
            'subject': scan['SUBJECT'],
            'session': scan['SESSION'],
            'scan': scan['SCANID']})

    return results


def _dicomzip2nifti(garjus, full_path):
    res = garjus.xnat().select(f'{full_path}/resources/DICOMZIP')

    files = res.files().get()

    if len(files) == 0:
        msg = f'no DICOMZIP files found:{full_path}'
        logger.info(msg)
        raise Exception(msg)
    elif len(files) > 1:
        msg = f'too many DICOMZIP files found:{full_path}'
        raise Exception(msg)

    src = files[0]

    with tempfile.TemporaryDirectory() as tmpdir:
        zip_path = os.path.join(tmpdir, src)
        res.file(src).get(zip_path)

        # unzip it
        unzipped_dir = pathlib.Path(f'{tmpdir}/UNZIPPED')
        unzipped_dir.mkdir()

        # Unzip the zip to the temp folder
        logger.debug(f'unzip {zip_path} to {unzipped_dir}')
        sb.run(['unzip', '-q', zip_path, '-d', unzipped_dir])

        # convert to NIFTI
        logger.info(f'convert to NIFTI:{full_path}')
        _d2n(unzipped_dir, res.parent())


def _dicomdir2nifti(garjus, full_path):
    res = garjus.xnat().select(f'{full_path}/resources/DICOM')

    files = res.files().get()

    if len(files) == 0:
        msg = f'no DICOM files found:{full_path}'
        logger.info(msg)
        raise Exception(msg)

    with tempfile.TemporaryDirectory() as tmpdir:
        res.get(tmpdir, extract=True)

        # convert to NIFTI
        logger.info(f'convert to NIFTI:{full_path}')
        _d2n(os.path.join(tmpdir, 'DICOM'), res.parent())


def _d2n(dicomdir, scan_object):
    nifti_list = []
    bval_path = ''
    bvec_path = ''
    json_path = ''

    # check that it hasn't been converted yet
    nifti_count = len(glob.glob(os.path.join(dicomdir, '*.nii.gz')))
    if nifti_count > 0:
        logger.info(f'nifti exists:{dicomdir}')
        return None

    # convert
    niftis = utils_dcm2nii.dicom2nifti(dicomdir)
    if not niftis:
        logger.info(f'nothing converted:{dicomdir}')
        return None

    # upload the converted files, NIFTI/JSON/BVAL/BVEC
    for fpath in glob.glob(os.path.join(dicomdir, '*')):
        if not os.path.isfile(fpath):
            continue

        if fpath.lower().endswith('.bval'):
            bval_path = utils_dcm2nii.sanitize_filename(fpath)
        elif fpath.lower().endswith('.bvec'):
            bvec_path = utils_dcm2nii.sanitize_filename(fpath)
        elif fpath.lower().endswith('.nii.gz'):
            nifti_list.append(utils_dcm2nii.sanitize_filename(fpath))
        elif fpath.lower().endswith('.json'):
            json_path = utils_dcm2nii.sanitize_filename(fpath)
        else:
            pass

    # More than one NIFTI
    if len(nifti_list) > 1:
        logger.warning('dcm2niix:multiple NIFTI')

    # Upload the NIFTIs
    logger.debug(f'uploading NIFTI:{nifti_list}')
    utils_xnat.upload_files(nifti_list, scan_object.resource('NIFTI'))

    if os.path.isfile(bval_path) and os.path.isfile(bvec_path):
        logger.debug('uploading BVAL/BVEC')
        utils_xnat.upload_file(bval_path, scan_object.resource('BVAL'))
        utils_xnat.upload_file(bvec_path, scan_object.resource('BVEC'))

    if os.path.isfile(json_path):
        logger.debug(f'uploading JSON:{json_path}')
        utils_xnat.upload_file(json_path, scan_object.resource('JSON'))



import time
import os

import pandas as pd
from dash import callback_context


def make_options(values):
    return [{'label': x, 'value': x} for x in values]


def make_columns(values):
    return [{'name': x, 'id': x} for x in values]


def read_data(filename):
    df = pd.read_pickle(filename)
    return df


def save_data(df, filename):
    df.to_pickle(filename)


def was_triggered(button_id):
    result = (
        callback_context.triggered and
        callback_context.triggered[0]['prop_id'].split('.')[0] == button_id)

    return result


def file_age(filename):
    return int((time.time() - os.path.getmtime(filename)) / 60)



import logging
import os
import time

from ..utils_redcap import upload_file, download_file


logger = logging.getLogger('garjus.automations.edat_convert2tab')


def process_project(
    project,
    events,
    raw_field,
    tab_field,
    convert_dir,
    wait_time=70,
    wait_count=3
):
    results = []
    def_field = project.def_field
    fields = [def_field, raw_field, tab_field]
    id2subj = {}

    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get mri records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event = r['redcap_event_name']

        try:
            subj = id2subj[record_id]
        except KeyError as err:
            logger.debug(f'record without secondary subject number:{err}')
            continue

        logger.debug(f'{subj}:{event}:{record_id}')

        # Check for existing converted file
        if r[tab_field]:
            logger.debug(f'already converted:{subj}:{event}')
            continue

        # Check for no raw file
        if not r[raw_field]:
            logger.debug(f'no raw edat file:{subj}:{event}')
            continue

        # Check for missing data flag
        if 'MISSING_DATA' in r[raw_field]:
            logger.debug(f'missing:{subj}:{event}')
            continue

        # Download to the convert dir
        basename = f'{subj}_{event}_{raw_field}.edat2'
        raw_file = f'{convert_dir}/{basename}'
        logger.debug(f'downloading {subj}:{event}:{raw_file}')

        res = download_file(
            project, record_id, raw_field, raw_file, event_id=event)
        if not res:
            logger.debug(f'dowload failed:{subj}:{event}')
            continue

        # here we wait for it to be converted and then try to find it
        found = False
        tab_file = f'{raw_file}_tab.txt'
        for i in range(wait_count):
            if os.path.exists(tab_file):
                found = True
                break

            logger.debug(f'{subj}:{event}:waiting for converted file:{i}')
            time.sleep(wait_time)

        if not found:
            logger.error(f'converted file not found:{subj}:{event}:{tab_file}')
            continue

        # Upload the converted file to redcap
        try:
            result = upload_file(
                project,
                record_id,
                tab_field,
                tab_file,
                event_id=event)
        except (ValueError) as err:
            logger.error(f'upload failed:{subj}:{event}:{tab_file}:{err}')

        if not result:
            logger.error(f'upload failed:{subj}:{event}')
            continue

        logger.info(f'uploaded:{subj}:{event}:{tab_file}')
        results.append({
            'result': 'COMPLETE',
            'description': 'edat_convert2tab',
            'category': 'edat_convert2tab',
            'subject': subj,
            'event': event,
            'field': tab_field})

    return results



"""Garjus Double Data Entry Comparison."""
from datetime import datetime
import tempfile
import logging
import os

from .dataentry_compare import run_compare


logger = logging.getLogger('garjus.compare')


def update(garjus, projects=None):
    """Update project progress."""
    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating double entry compare:{p}')
            update_project(garjus, p)


def update_project(garjus, project):
    """Update project double entry."""
    double_reports = garjus.double_reports(project)

    # what time is it? use this for naming
    now = datetime.now()

    # determine curxrent month and year to get current monthly repot id
    cur_double = now.strftime("%B%Y")

    # check that each project has report for current month
    has_cur = any(d.get('double_name') == cur_double for d in double_reports)
    if not has_cur:
        logger.debug(f'making new double record:{project}:{cur_double}')
        make_double(garjus, project, cur_double, now)
    else:
        logger.debug(f'double entry record exists:{project}:{cur_double}')


def make_double(garjus, project, cur_double, now):
    # Get the projects to compare
    proj_primary = garjus.primary(project)
    proj_secondary = garjus.secondary(project)

    if not proj_primary:
        logger.debug(f'cannot compare, primary REDCap not set:{project}')
        return

    if not proj_secondary:
        logger.debug(f'cannot run, secondary REDCap not set:{project}')
        return

    """Make double entry comparison report."""
    with tempfile.TemporaryDirectory() as outdir:
        logger.debug(f'created temporary directory:{outdir}')

        fnow = now.strftime("%Y-%m-%d_%H_%M_%S")
        pdf_file = f'{outdir}/{project}_report_{fnow}.pdf'
        excel_file = f'{outdir}/{project}_results_{fnow}.xlsx'

        logger.debug(f'making report:{pdf_file}')
        make_double_report(proj_primary, proj_secondary, pdf_file, excel_file)

        if not os.path.isfile(pdf_file):
            logger.debug(f'no results')
        else:
            logger.debug(f'uploading results')
            garjus.add_double(project, cur_double, now, pdf_file, excel_file)


def make_double_report(proj_primary, proj_secondary, pdf_file, excel_file):
    # Run it
    p1 = proj_primary.export_project_info().get('project_title')
    p2 = proj_secondary.export_project_info().get('project_title')
    logger.debug(f'compare {p1} to {p2}')
    run_compare(proj_primary, proj_secondary, pdf_file, excel_file)



import logging
import tempfile

import pandas as pd

from ...utils_redcap import get_redcap, download_file, field2events


logger = logging.getLogger('garjus.automations')


reg_field = 'toolbox_regdata'
score_field = 'toolbox_cogscores'
done_field = 'toolbox_pin'


def run(project, record_id, event_id):
    data = None

    with tempfile.TemporaryDirectory() as tmpdir:
        reg_file = f'{tmpdir}/regfile.csv'
        score_file = f'{tmpdir}/scorefile.csv'

        # Download files from redcap
        logger.debug(f'downloading file:{record_id}:{event_id}:{reg_field}:{reg_file}')
        download_file(project, record_id, reg_field, reg_file, event_id=event_id)
        logger.debug(f'downloading file:{record_id}:{event_id}:{score_field}:{score_file}')
        download_file(project, record_id, score_field, score_file, event_id=event_id)

        # Extract data from downloaded files
        reg_data = toolbox_extract_regdata(reg_file)

        score_data = toolbox_extract_cogscores(score_file)

        # Transform data to match redcap field names
        data = toolbox_transform(reg_data, score_data)

    # Load data back to redcap
    toolbox_load(project, record_id, event_id, data)


def toolbox_transform(regdata, scoredata):
    # Initialize test data
    picseqtest = None
    listsorttest = None
    patterntest = None
    picvocabtest = None
    oralrecogtest = None
    cogcrystalcomp = None
    data = {}

    # Start with the registration data
    data.update({
        'toolbox_pin': regdata['PIN'],
        'toolbox_deviceid': regdata['DeviceID'],
        'toolbox_age': regdata['Age'],
        'toolbox_education': regdata['Education'],
        'toolbox_gender': regdata['Gender'],
        'toolbox_handedness': regdata['Handedness'],
        'toolbox_race': regdata['Race'],
        'toolbox_ethnicity': regdata['Ethnicity'],
        'toolbox_assessment': regdata['Assessment Name'],
    })

    # Find the Pic Seq data that has mutliple versions
    for i in list(scoredata.keys()):
        if i.startswith('NIH Toolbox Picture Sequence Memory Test'):
            picseqtest = scoredata[i]

    # Load the other tests
    listsorttest = scoredata['NIH Toolbox List Sorting Working Memory Test Age 7+ v2.1']
    patterntest = scoredata['NIH Toolbox Pattern Comparison Processing Speed Test Age 7+ v2.1']
    picvocabtest = scoredata['NIH Toolbox Picture Vocabulary Test Age 3+ v2.1']
    oralrecogtest = scoredata['NIH Toolbox Oral Reading Recognition Test Age 3+ v2.1']
    flankertest = scoredata['NIH Toolbox Flanker Inhibitory Control and Attention Test Age 12+ v2.1']
    cardsorttest = scoredata['NIH Toolbox Dimensional Change Card Sort Test Age 12+ v2.1']

    # Get the individual scores
    data.update({
        'toolbox_listsorttest_raw': listsorttest['RawScore'],
        'toolbox_patterntest_raw': patterntest['RawScore'],
        'toolbox_picseqtest_raw': picseqtest['RawScore'],
        'toolbox_oralrecogtest_theta': oralrecogtest['Theta'],
        'toolbox_picseqtest_theta': picseqtest['Theta'],
        'toolbox_picvocabtest_theta': picvocabtest['Theta'],
        'toolbox_listsorttest_uncstd': listsorttest['Uncorrected Standard Score'],
        'toolbox_oralrecogtest_uncstd': oralrecogtest['Uncorrected Standard Score'],
        'toolbox_patterntest_uncstd': patterntest['Uncorrected Standard Score'],
        'toolbox_picseqtest_uncstd': picseqtest['Uncorrected Standard Score'],
        'toolbox_picvocabtest_uncstd': picvocabtest['Uncorrected Standard Score'],
        'toolbox_listsorttest_agestd': listsorttest['Age-Corrected Standard Score'],
        'toolbox_oralrecogtest_agestd': oralrecogtest['Age-Corrected Standard Score'],
        'toolbox_patterntest_agestd': patterntest['Age-Corrected Standard Score'],
        'toolbox_picseqtest_agestd': picseqtest['Age-Corrected Standard Score'],
        'toolbox_picvocabtest_agestd': picvocabtest['Age-Corrected Standard Score'],
        'toolbox_listsorttest_tscore': listsorttest['Fully-Corrected T-score'],
        'toolbox_oralrecogtest_tscore': oralrecogtest['Fully-Corrected T-score'],
        'toolbox_patterntest_tscore': patterntest['Fully-Corrected T-score'],
        'toolbox_picseqtest_tscore': picseqtest['Fully-Corrected T-score'],
        'toolbox_picvocabtest_tscore': picvocabtest['Fully-Corrected T-score'],
        'toolbox_flankertest_raw': flankertest['RawScore'],
        'toolbox_flankertest_uncstd': flankertest['Uncorrected Standard Score'],
        'toolbox_flankertest_agestd': flankertest['Age-Corrected Standard Score'],
        'toolbox_flankertest_tscore': flankertest['Fully-Corrected T-score'],
        'toolbox_cardsorttest_raw': cardsorttest['RawScore'],
        'toolbox_cardsorttest_uncstd': cardsorttest['Uncorrected Standard Score'],
        'toolbox_cardsorttest_agestd': cardsorttest['Age-Corrected Standard Score'],
        'toolbox_cardsorttest_tscore': cardsorttest['Fully-Corrected T-score'],
    })

    cogcrystalcomp = scoredata.get('Cognition Crystallized Composite v1.1', None)
    cogfluidcomp = scoredata.get('Cognition Fluid Composite v1.1', None)
    cogearlycomp = scoredata.get('Cognition Early Childhood Composite v1.1', None)
    cogtotalcomp = scoredata.get('Cognition Total Composite Score v1.1', None)
    audlearntest = scoredata.get('NIH Toolbox Auditory Verbal Learning Test (Rey) Age 8+ v2.0', None)

    if audlearntest:
        data.update({
            'toolbox_audlearntest_raw': audlearntest['RawScore'],
        })

    if cogcrystalcomp:
        data.update({
            'toolbox_cogcrystalcomp_uncstd': cogcrystalcomp['Uncorrected Standard Score'],
            'toolbox_cogcrystalcomp_agestd': cogcrystalcomp['Age-Corrected Standard Score'],
            'toolbox_cogcrystalcomp_tscore': cogcrystalcomp['Fully-Corrected T-score'],
        })

    if cogfluidcomp:
        data.update({
            'toolbox_cogfluidcomp_uncstd': cogfluidcomp['Uncorrected Standard Score'],
            'toolbox_cogfluidcomp_agestd': cogfluidcomp['Age-Corrected Standard Score'],
            'toolbox_cogfluidcomp_tscore': cogfluidcomp['Fully-Corrected T-score'],
        })

    if cogearlycomp:
        data.update({
            'toolbox_cogearlycomp_uncstd': cogearlycomp['Uncorrected Standard Score'],
            'toolbox_cogearlycomp_agestd': cogearlycomp['Age-Corrected Standard Score'],
            'toolbox_cogearlycomp_tscore': cogearlycomp['Fully-Corrected T-score'],
        })

    if cogearlycomp:
        data.update({
            'toolbox_cogtotalcomp_uncstd': cogtotalcomp['Uncorrected Standard Score'],
            'toolbox_cogtotalcomp_agestd': cogtotalcomp['Age-Corrected Standard Score'],
            'toolbox_cogtotalcomp_tscore': cogtotalcomp['Fully-Corrected T-score'],
        })

    return data


def toolbox_load(project, record_id, event_id, data):
    data[project.def_field] = record_id
    data['redcap_event_name'] = event_id

    try:
        response = project.import_records([data])
        assert 'count' in response
        logger.debug('uploaded')
    except AssertionError as e:
        logger.error('error uploading', record_id, e)


def toolbox_extract_regdata(filename):
    data = {}

    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Get data from last row
    data = df.iloc[-1].to_dict()

    return data


def toolbox_extract_cogscores(filename):
    data = {}

    # Load csv
    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Drop instrument duplicates, keeping the last only
    df = df.drop_duplicates(subset='Inst', keep='last')

    # convert to dict of dicts indexed by Instrument
    df = df.dropna(subset=['Inst'])
    df = df.set_index('Inst')
    data = df.to_dict('index')

    return data


def process_project(project):
    results = []
    events = field2events(project, reg_field)

    records = project.export_records(
        fields=[project.def_field, done_field, reg_field, score_field],
        events=events)

    for r in records:
        record_id = r[project.def_field]
        event_id = r['redcap_event_name']

        if r[done_field]:
            logger.debug(f'already ETL:{record_id}:{event_id}')
            continue

        if not r[score_field]:
            logger.debug(f'no data file:{record_id}:{event_id}')
            continue

        logger.debug(f'running ETL:{record_id}:{event_id}')
        results.append({'subject': record_id, 'event': event_id})
        run(project, record_id, event_id)

    return results


if __name__ == "__main__":
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    process_project(get_redcap('', api_key=''))
    print('Done!')



import logging
import tempfile
import io

import pandas as pd
import numpy as np

from ...utils_redcap import field2events, download_file


def parse_tat(filename):
    encoding = 'utf-8'
    skiprows = 0
    first_field = 'ExperimentName'

    # Determine how many rows to skip prior to header
    with io.open(filename, encoding=encoding) as f:
        for line in f:
            if line.startswith(first_field):
                break
            else:
                skiprows += 1

    # Load Data
    df = pd.read_table(
        filename, sep='\t', encoding=encoding, skiprows=skiprows, header=0)

    # First exclude the practice
    df = df[df['Procedure'] == 'TrialProc']

    # Combine respones from Target and Mask
    df = df.apply(combine_tat_responses, axis=1)

    # Drop any rows without a response
    df = df.dropna(subset=['Combined.RESP'])

    return df


def combine_tat_responses(row):
    if not pd.isnull(row['Target.RESP']):
        row['Combined.RESP'] = row['Target.RESP']
        row['Combined.RT'] = row['Target.RT']
    elif not pd.isnull(row['Mask.RESP']):
        row['Combined.RESP'] = row['Mask.RESP']
        row['Combined.RT'] = row['Mask.RT']
    else:
        row['Combined.RESP'] = np.nan
        row['Combined.RT'] = np.nan

    return row


def load_tatfile(filename):
    df = parse_tat(filename)
    data = {
        'total_tm': '0',
        'number_endorsed_bad_tm': '0',
        'number_reject_bad_tm': '0',
        'number_endorsed_good_tm': '0',
        'number_reject_good_tm': '0',
        'rt_endorsed_bad_tm': '',
        'rt_reject_bad_tm': '',
        'rt_endorsed_good_tm': '',
        'rt_reject_good_tm': ''}

    if len(df) > 0:
        data['total_tm'] = len(df)

        # Remove whitespace that we know is there
        df['EmoWord'] = df['EmoWord'].str.strip()

        # Separate good and bad words
        dfb = df[df['EmoWord'] == 'bad']
        dfg = df[df['EmoWord'] == 'good']

        if len(dfb) > 0:
            data['number_endorsed_bad_tm'] = len(dfb[dfb['Combined.RESP'] == 'm'])
            data['rt_endorsed_bad_tm'] = np.around(np.mean(dfb[dfb['Combined.RESP'] == 'm']['Combined.RT']), decimals=1)
            data['number_reject_bad_tm'] = len(dfb[dfb['Combined.RESP'] == 'x'])
            data['rt_reject_bad_tm'] = np.around(np.mean(dfb[dfb['Combined.RESP'] == 'x']['Combined.RT']), decimals=1)

        if len(dfg) > 0:
            data['number_endorsed_good_tm'] = len(dfg[dfg['Combined.RESP'] == 'm'])
            data['rt_endorsed_good_tm'] = np.around(np.mean(dfg[dfg['Combined.RESP'] == 'm']['Combined.RT']), decimals=1)
            data['number_reject_good_tm'] = len(dfg[dfg['Combined.RESP'] == 'x'])
            data['rt_reject_good_tm'] = np.around(np.mean(dfg[dfg['Combined.RESP'] == 'x']['Combined.RT']), decimals=1)

    return data


def etl_tat(project, record_id, event_id):
    data = None
    tab_file = None
    tab_field = 'tat_conv'

    with tempfile.TemporaryDirectory() as tmpdir:
        tab_file = f'{tmpdir}/{record_id}-{event_id}-{tab_field}.txt'

        # Download TAT converted tab file
        try:
            download_file(
                project,
                record_id,
                tab_field,
                tab_file,
                event_id=event_id)
        except Exception as err:
            logging.error(f'download failed:{record_id}:{event_id}:{err}')
            return False

        # Transform data
        try:
            data = _transform(tab_file)
            data[project.def_field] = record_id
            data['redcap_event_name'] = event_id
        except Exception as err:
            logging.error(f'transform failed:{record_id}:{event_id}:{err}')
            return False

    # Load the data back to redcap
    try:
        _response = project.import_records([data])
        assert 'count' in _response
        logging.info(f'TAT uploaded:{record_id}')
    except AssertionError as err:
        logging.error(f'TAT upload:{record_id}:{err}')
        return False

    return True


def _transform(filename):
    # Load the data
    logging.info('{}:{}'.format('extracting tat', filename))
    tat_data = load_tatfile(filename)
    if tat_data is None:
        logging.error('extract failed')
        return

    # Transform the data
    data = {}
    data['total_tm'] = str(tat_data['total_tm'])
    data['number_endorsed_good_tm'] = str(tat_data['number_endorsed_good_tm'])
    data['number_endorsed_bad_tm'] = str(tat_data['number_endorsed_bad_tm'])
    data['number_reject_good_tm'] = str(tat_data['number_reject_good_tm'])
    data['number_rejected_bad_tm'] = str(tat_data['number_reject_bad_tm'])
    data['rt_endorsed_good_tm'] = str(tat_data['rt_endorsed_good_tm'])
    data['rt_endorsed_bad_tm'] = str(tat_data['rt_endorsed_bad_tm'])
    data['rt_reject_good_tm'] = str(tat_data['rt_reject_good_tm'])
    data['rt_reject_bad_tm'] = str(tat_data['rt_reject_bad_tm'])

    return data


def process_project(project):
    tab_field  = 'tat_conv'
    done_field = 'total_tm'
    results = []
    def_field = project.def_field
    fields = [def_field, tab_field, done_field]
    id2subj = {}
    events = field2events(project, tab_field)
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event_id = r['redcap_event_name']
        subj = id2subj.get(record_id, record_id)

        # Check for converted file
        if not r[tab_field]:
            logging.debug(f'not yet converted:{record_id}:{subj}:{event_id}')
            continue

        # Determine if ETL has already been run
        if r[done_field]:
            logging.debug(f'already ETL:{record_id}:{subj}:{event_id}')
            continue

        # Do the ETL
        result = etl_tat(project, record_id, event_id)
        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_traitadjtask',
                'description': 'etl_traitadjtask',
                'subject': subj,
                'event': event_id,
                'field': tab_field})

    return results



'''Subjects from REDCap.'''
import logging

import pandas as pd


logger = logging.getLogger('garjus.subjects')


def load_subjects(garjus, project, include_dob=False):
    project_redcap = garjus.primary(project)

    if not project_redcap:
        return pd.DataFrame([], columns=['ID', 'PROJECT'])

    def_field = project_redcap.def_field
    sec_field = project_redcap.export_project_info()['secondary_unique_field']
    guid_field = None
    sex_field = None
    dob_field = None
    date_field = None
    field_names = project_redcap.field_names

    if 'guid' in field_names:
        guid_field = 'guid'

    if 'dob' in field_names:
        dob_field = 'dob'
    elif 'dob_sub' in field_names:
        dob_field = 'dob_sub'

    if 'sex_xcount' in field_names:
        sex_field = 'sex_xcount'
    elif 'dems_sex' in field_names:
        sex_field = 'dems_sex'
    elif 'sex_demo' in field_names:
        sex_field = 'sex_demo'

    if 'mri_date' in field_names:
        date_field = 'mri_date'

    # Load subject records from redcap
    fields = [def_field]

    if sec_field:
        fields.append(sec_field)

    if guid_field:
        fields.append(guid_field)

    if dob_field:
        fields.append(dob_field)

    if sex_field:
        fields.append(sex_field)

    rec = project_redcap.export_records(fields=fields, raw_or_label='label')

    # Ignore records without secondary ID
    if sec_field:
        rec = [x for x in rec if x[sec_field]]

    # Must have dob to calc age
    if dob_field:
        try:
            rec = [x for x in rec if x[dob_field]]
        except KeyError as err:
            logger.debug(f'cannot access dob:{dob_field}:{err}')

    # Make data frame
    if project_redcap.is_longitudinal:
        df = pd.DataFrame(rec, columns=fields + ['redcap_event_name'])
    else:
        df = pd.DataFrame(rec, columns=fields)

    # Set the project
    df['PROJECT'] = project

    # Determine group
    df['GROUP'] = 'UNKNOWN'

    if project == 'DepMIND2':
        # All DM2 are depressed
        df['GROUP'] = 'Depress'
    elif project == 'D3':
        # Use arm/events names to determine group
        df['GROUP'] = df['redcap_event_name'].map({
            'Screening (Arm 2: Never Depressed)': 'Control',
            'Screening (Arm 1: Currently Depressed)': 'Depress',
        })
    elif project == 'REMBRANDT':
        # Use arm/events names to determine which arm
        df['GROUP'] = df['redcap_event_name'].map({
            'Screening (Arm 3: Longitudinal Phase: Remitted)': 'Depress',
            'Screening (Arm 2: Longitudinal Phase: Never Depressed)': 'Control',
        })

    # Load MRI records to get first date
    if dob_field and date_field:
        fields = [def_field, date_field]
        rec = project_redcap.export_records(fields=fields, raw_or_label='label')
        rec = [x for x in rec if x[date_field]]
        dfm = pd.DataFrame(rec, columns=fields)
        dfm = dfm.astype(str)
        dfm = dfm.sort_values(date_field)
        dfm = dfm.drop_duplicates(subset=[def_field], keep='first')

        # Merge in date
        df = pd.merge(df, dfm, how='left', on=def_field)

        # Exclude incomplete data
        df = df.dropna()

        # Calculate age at baseline
        df[dob_field] = pd.to_datetime(df[dob_field])
        df[date_field] = pd.to_datetime(df[date_field])
        df['AGE'] = (
            df[date_field] - df[dob_field]
        ).values.astype('<m8[Y]').astype('int').astype('str')

        if include_dob:
            df['DOB'] = df[dob_field]

    # Exclude incomplete data
    df = df.dropna()

    if sex_field:
        df['SEX'] = df[sex_field].map({'Male': 'M', 'Female': 'F'})

    if guid_field:
        df['GUID'] = df[guid_field]

    if sec_field:
        df['ID'] = df[sec_field]
    else:
        df['ID'] = df[def_field]

    # Drop intermediate columns
    drop_columns = [def_field, 'redcap_event_name', dob_field, date_field, sec_field, guid_field, sex_field]
    drop_columns = [x for x in drop_columns if x and x in df.columns]
    df = df.drop(columns=drop_columns)

    # Finish up
    df = df.sort_values('ID')
    df = df.drop_duplicates()
    df = df.set_index('ID')

    return df



"""qa dashboard tab.

# DESCRIPTION:
# the table is by session using a pivottable that aggregates the statuses
# for each scan/assr type. then we have dropdowns to filter by project,
# processing type, scan type, etc.
"""

# TODO: connect sessions from same subject baseline with arc line or something
# or could have "by subject" choice that has a subject per y value?

# TODO: time dropdown to select from: last week, last month, this month,
# could also just be separate buttons beside the date picker

# TODO: checkboxes to hide columns DATE, SESSTYPE, SITE, NOTE: often blank


import logging
import re
import itertools

import pandas as pd
import numpy as np
import plotly
import plotly.graph_objs as go
import plotly.subplots
from dash import dcc, html, dash_table as dt
from dash import Input, Output
import dash_bootstrap_components as dbc

from ..app import app
from .. import utils
from ..shared import QASTATUS2COLOR, RGB_DKBLUE, GWIDTH
from . import data
from ...garjus import Garjus


logger = logging.getLogger('dashboard.qa')


LEGEND1 = '''
✅QA Passed ㅤ
🟩QA TBD ㅤ
❌QA Failed ㅤ
🩷Job Failed ㅤ
🟡Needs Inputs ㅤ
🔷Job Running
□ None Found
'''

LEGEND2 = '''
🧠 MR
☢️ PET
🤯 EEG
'''

MOD2EMO = {'MR': '🧠', 'PET': '☢️', 'EEG': '🤯'}


# The data will be pivoted by session to show a row per session and
# a column per scan/assessor type,
# the values in the column a string of characters
# that represent the status of one scan or assesor,
# the number of characters is the number of scans or assessors
# the columns will be the merged
# status column with harmonized values to be red/yellow/green/blue


def _get_graph_content(dfp):
    tabs_content = []

    logger.debug('get_qa_figure')

    # Check for empty data
    if dfp is None or len(dfp) == 0:
        logger.debug('empty data, using empty figure')
        return [html.Div(html.H1('Choose Project(s) to load'))]

    # Make a 1x1 figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    # First we copy the dfp and then replace the values in each
    # scan/proc type column with a metastatus,
    # that gives us a high level status of the type for that session

    # TODO: should we just make a different pivot table here going back to
    # the original df? yes, later
    dfp_copy = dfp.copy()
    for col in dfp_copy.columns:
        if col in ('SESSION', 'PROJECT', 'DATE', 'NOTE'):
            # don't mess with these columns
            # TODO: do we need this if we haven't reindexed yet?
            continue

        # Change each value from the multiple values in concatenated
        # characters to a single overall status
        dfp_copy[col] = dfp_copy[col].apply(get_metastatus)

    # The pivot table for the graph is a pivot of the pivot table, instead
    # of having a row per session, this pivot table has a row per
    # pivot_type, we can pivot by type to get counts of each status for each
    # scan/proc type, or we can pivot by project to get counts of sessions
    # for each project
    # The result will be a table with one row per TYPE (index=TYPE),
    # and we'll have a column for each STATUS (so columns=STATUS),
    # and we'll count how many sessions (values='SESSION') we find for each
    # cell get a copy so it's defragmented
    dfp_copy = dfp_copy.reset_index().copy()

    # don't need subject
    dfp_copy = dfp_copy.drop(columns=['SUBJECT', 'SUBJECTLINK'])

    # use pandas melt function to unpivot our pivot table
    df = pd.melt(
        dfp_copy,
        id_vars=(
            'SESSION',
            'SESSIONLINK',
            'PROJECT',
            'DATE',
            'SITE',
            'SESSTYPE',
            'MODALITY',
            'NOTE'),
        value_name='STATUS')

    # We use fill_value to replace nan with 0
    dfpp = df.pivot_table(
        index='TYPE',
        columns='STATUS',
        values='SESSION',
        aggfunc='count',
        fill_value=0)

    # sort so scans are first, then assessor
    scan_type = []
    assr_type = []
    for cur_type in dfpp.index:
        # Use a regex to test if name ends with _v and a number, then assr
        if re.search('_v\d+$', cur_type):
            assr_type.append(cur_type)
        else:
            scan_type.append(cur_type)

    newindex = scan_type + assr_type
    dfpp = dfpp.reindex(index=newindex)

    # Draw bar for each status, these will be displayed in order
    # ydata should be the types, xdata should be count of status
    # for each type
    for cur_status, cur_color in QASTATUS2COLOR.items():
        ydata = dfpp.index
        if cur_status not in dfpp:
            xdata = [0] * len(dfpp.index)
        else:
            xdata = dfpp[cur_status]

        cur_name = '{} ({})'.format(cur_status, sum(xdata))

        fig.append_trace(
            go.Bar(
                x=ydata,
                y=xdata,
                name=cur_name,
                marker=dict(color=cur_color),
                opacity=0.9),
            1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=True, width=GWIDTH)

    # Build the tab
    label = 'By {}'.format('TYPE')
    graph = html.Div(dcc.Graph(figure=fig))
    tabs_content.append(dbc.Tab(label=label, children=[graph]))

    # We also want a tab for By Project, so we can ask e.g. how many
    # sessions for each project, and then ask
    # which projects have a T1 and a good FS6_v1
    # later combine with other pivot
    # table and loop on pivot type
    dfpp = df.pivot_table(
        index='PROJECT',
        values='SESSION',
        aggfunc=pd.Series.nunique,
        fill_value=0)

    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    ydata = dfpp.index
    xdata = dfpp.SESSION

    cur_name = '{} ({})'.format('ALL', sum(xdata))
    cur_color = RGB_DKBLUE

    fig.append_trace(
        go.Bar(
            x=ydata,
            y=xdata,
            text=xdata,
            name=cur_name,
            marker=dict(color=cur_color),
            opacity=0.9),
        1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=True, width=GWIDTH)

    # Build the tab
    label = 'By {}'.format('PROJECT')
    graph = html.Div(
        dcc.Graph(figure=fig),
        style={'width': '100%', 'display': 'inline-block'}
    )
    tabs_content.append(dbc.Tab(label=label, children=[graph]))

    # Append the by-time graph (this was added later with separate function)
    dfs = df[['PROJECT', 'DATE', 'SESSION', 'SESSTYPE', 'SITE', 'MODALITY']].drop_duplicates()
    fig = _sessionsbytime_figure(dfs, selected_groupby='PROJECT')
    label = 'By {}'.format('TIME')
    graph = html.Div(dcc.Graph(figure=fig), style={
        'width': '100%', 'display': 'inline-block'})
    tabs_content.append(dbc.Tab(label=label, children=[graph]))

    # Return the tabs wrapped in a spinning loader
    return [dbc.Tabs(id='tabs-qa', children=tabs_content)]


def _sessionsbytime_figure(df, selected_groupby):
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    from itertools import cycle
    import plotly.express as px
    palette = cycle(px.colors.qualitative.Plotly)

    for mod, sesstype in itertools.product(df.MODALITY.unique(), df.SESSTYPE.unique()):

        # Get subset for this session type
        dfs = df[(df.SESSTYPE == sesstype) & (df.MODALITY == mod)]

        # Nothing to plot so go to next session type
        if dfs.empty:
            continue

        # Create boxplot for this var and add to figure
        # Default to the jittered boxplot with no boxes

        # markers symbols, see https://plotly.com/python/marker-style/
        if mod == 'MR':
            symb = 'circle-dot'
        elif mod == 'PET':
            symb = 'diamond-wide-dot'
        else:
            symb = 'diamond-tall-dot'

        _color = next(palette)

        # Convert hex to rgba with alpha of 0.5
        if _color.startswith('#'):
            _rgba = 'rgba({},{},{},{})'.format(
                int(_color[1:3], 16),
                int(_color[3:5], 16),
                int(_color[5:7], 16),
                0.7)
        else:
            _r, _g, _b = _color[4:-1].split(',')
            _a = 0.7
            _rgba = 'rgba({},{},{},{})'.format(_r, _g, _b, _a)

        # Plot this session type
        _row = 1
        _col = 1
        fig.append_trace(
            go.Box(
                name='{} {} ({})'.format(sesstype, mod, len(dfs)),
                x=dfs['DATE'],
                y=dfs[selected_groupby],
                boxpoints='all',
                jitter=0.7,
                text=dfs['SESSION'],
                pointpos=0.5,
                orientation='h',
                marker={
                    'symbol': symb,
                    'color': _rgba,
                    'size': 12,
                    'line': dict(width=2, color=_color)
                },
                line={'color': 'rgba(0,0,0,0)'},
                fillcolor='rgba(0,0,0,0)',
                hoveron='points',
            ),
            _row,
            _col)

        # show lines so we can better distinguish categories
        fig.update_yaxes(showgrid=True)

        x_mins = []
        x_maxs = []
        for trace_data in fig.data:
            x_mins.append(min(trace_data.x))
            x_maxs.append(max(trace_data.x))

        fig.update_layout(width=GWIDTH)

    return fig


def get_content():
    '''Get QA page content.'''

    # We use the dbc grid layout with rows and columns, rows are 12 units wide
    content = [
        dbc.Row(html.Div(id='div-qa-graph', children=[])),
        dbc.Row([
            dbc.Col(
                dcc.DatePickerRange(id='dpr-qa-time', clearable=True),
                width=5,
            ),
            dbc.Col(
                dbc.Button(
                    'Refresh Data',
                    outline=True,
                    className="me-1",
                    id='button-qa-refresh',
                    size='sm',
                    color='primary',
                ),
                align='center',
            ),
            dbc.Col(
                dbc.Switch(
                    id='switch-qa-autofilter',
                    label='Autofilter',
                    value=True,
                ),
                align='center',
            ),
            dbc.Col(
                dbc.Switch(
                    id='switch-qa-graph',
                    label='Graph',
                    value=False,
                ),
                align='center',
            ),
        ]),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-qa-proj',
                    multi=True,
                    placeholder='Select Project(s)',
                    value=[],
                ),
                width=5,
            ),
        ),
        dbc.Row([
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-qa-sess',
                    multi=True,
                    placeholder='Select Session Type(s)',
                ),
                width=5,
            ),
            dbc.Col(
                dbc.Checklist(
                    options=[
                        {'label': '🧠 MR', 'value': 'MR'},
                        {'label': '☢️ PET', 'value': 'PET'},
                        {'label': '🤯 EEG', 'value': 'EEG'},
                    ],
                    value=['MR', 'PET', 'EEG'],
                    id='switches-qa-modality',
                    inline=True,
                    switch=True
                ),
                align='center',
            ),
        ]),
        dbc.Row([
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-qa-proc',
                    multi=True,
                    placeholder='Select Processing Type(s)',
                ),
                width=5,
            ),
            dbc.Col(
                dbc.Checklist(
                    options=[
                        {'label': '✅', 'value': 'P'},
                        {'label': '🟩', 'value': 'Q'},
                        {'label': '❌', 'value': 'F'},
                        {'label': '🩷', 'value': 'X'},
                        {'label': '🟡', 'value': 'N'},
                        {'label': '🔷', 'value': 'R'},
                        {'label': '□', 'value': 'E'},
                    ],
                    value=['P', 'Q', 'F', 'X', 'N', 'R', 'E'],
                    id='switches-qa-procstatus',
                    inline=True,
                    switch=True
                ),
                align='center',
            ),
        ]),
        dbc.Row([
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-qa-scan',
                    multi=True,
                    placeholder='Select Scan Type(s)',
                ),
                width=5,
            ),
        ]),
        dbc.Row(
            [
                dbc.Col(
                    dbc.RadioItems(
                        # Use specific css to make radios look like buttons
                        className="btn-group",
                        inputClassName="btn-check",
                        labelClassName="btn btn-outline-primary",
                        labelCheckedClassName="active",
                        options=[
                            {'label': 'Sessions', 'value': 'sess'},
                            {'label': 'Subjects', 'value': 'subj'},
                            {'label': 'Projects', 'value': 'proj'},
                        ],
                        value='sess',
                        id='radio-qa-pivot',
                    ),
                    align='end',
                    width=5,
                ),
            ],
            align='end',
        ),
        dbc.Spinner(id="loading-qa-table", children=[
            dbc.Label('Get ready...', id='label-qa-rowcount1'),
        ]),
        dt.DataTable(
            columns=[],
            data=[],
            filter_action='native',
            page_action='none',
            sort_action='native',
            id='datatable-qa',
            style_table={
                'overflowY': 'scroll',
                'overflowX': 'scroll',
            },
            style_cell={
                'textAlign': 'center',
                'padding': '5px 5px 0px 5px',
                'width': '30px',
                'overflow': 'hidden',
                'textOverflow': 'ellipsis',
                'height': 'auto',
                'minWidth': '40',
                'maxWidth': '70'
            },
            style_header={
                'fontWeight': 'bold',
                'padding': '5px 15px 0px 10px',
            },
            style_cell_conditional=[
                {'if': {'column_id': 'NOTE'}, 'textAlign': 'left'},
                {'if': {'column_id': 'SESSIONS'}, 'textAlign': 'left'},
                {'if': {'column_id': 'SESSION'}, 'textAlign': 'center'},
            ],
            css=[dict(selector="p", rule="margin: 0; text-align: center")],
            fill_width=False,
            export_format='xlsx',
            export_headers='names',
            export_columns='visible'
        ),
        dbc.Label('Get ready...', id='label-qa-rowcount2'),
        html.Div([
            html.P(
                LEGEND1,
                style={'marginTop': '15px', 'textAlign': 'center'}
            ),
            html.P(
                LEGEND2,
                style={'textAlign': 'center'}
            )],
            style={'textAlign': 'center'},
        ),
    ]

    return content


def get_metastatus(status):

    if status != status:
        # empty so it's none
        metastatus = 'NONE'
    elif not status or pd.isnull(status):
        # empty so it's none
        metastatus = 'NONE'
    elif 'P' in status:
        # at least one passed, so PASSED
        metastatus = 'PASS'
    elif 'Q' in status:
        # any are still needs qa, then 'NEEDS_QA'
        metastatus = 'NQA'
    elif 'N' in status:
        # if any jobs are still running, then NEEDS INPUTS?
        metastatus = 'NPUT'
    elif 'F' in status:
        # at this point if one failed, then they all failed, so 'FAILED'
        metastatus = 'FAIL'
    elif 'X' in status:
        metastatus = 'JOBF'
    elif 'R' in status:
        metastatus = 'JOBR'
    else:
        # whatever else is UNKNOWN, grey
        metastatus = 'NONE'

    return metastatus


def qa_pivot(df):
    df.DATE = df.DATE.fillna('')

    dfp = df.pivot_table(
        index=(
            'SESSION', 'SESSIONLINK', 'SUBJECT', 'SUBJECTLINK', 'PROJECT',
            'DATE', 'SESSTYPE', 'SITE', 'MODALITY', 'NOTE'),
        columns='TYPE',
        values='STATUS',
        aggfunc=lambda x: ''.join(x))

    # and return our pivot table
    return dfp


# This is where the data gets initialized
def load_data(projects=[], refresh=False, hidetypes=True):
    if projects is None:
        projects = []

    return data.load_data(
        projects=projects,
        refresh=refresh,
        hidetypes=hidetypes)


def load_options(df):
    garjus = Garjus()
    projects = garjus.projects()
    sesstypes = []
    proctypes = []
    scantypes = []

    # Filter to selected
    scantypes = df.SCANTYPE.unique()

    # Remove blanks and sort
    scantypes = [x for x in scantypes if x]
    scantypes = sorted(scantypes)

    # Now sessions
    sesstypes = df.SESSTYPE.unique()
    sesstypes = [x for x in sesstypes if x]
    sesstypes = sorted(sesstypes)

    # And finally proc
    proctypes = df.PROCTYPE.unique()
    proctypes = [x for x in proctypes if x]
    proctypes = sorted(proctypes)

    return projects, sesstypes, proctypes, scantypes


# Initialize the callbacks for the app

# inputs:
# values from assr proc types dropdown
# values from project dropdown
# values from timeframe dropdown
# number of clicks on refresh button

# returns:
# options for the assessor proc types dropdown
# options for the assessor projects dropdown
# options for the assessor scans dropdown
# options for the assessor sessions dropdown
# data for the table
# content for the graph tabs
@app.callback(
    [Output('dropdown-qa-proc', 'options'),
     Output('dropdown-qa-scan', 'options'),
     Output('dropdown-qa-sess', 'options'),
     Output('dropdown-qa-proj', 'options'),
     Output('datatable-qa', 'data'),
     Output('datatable-qa', 'columns'),
     Output('div-qa-graph', 'children'),
     Output('label-qa-rowcount1', 'children'),
     Output('label-qa-rowcount2', 'children'),
     ],
    [Input('dropdown-qa-proc', 'value'),
     Input('dropdown-qa-scan', 'value'),
     Input('dropdown-qa-sess', 'value'),
     Input('dropdown-qa-proj', 'value'),
     Input('dpr-qa-time', 'start_date'),
     Input('dpr-qa-time', 'end_date'),
     Input('switch-qa-autofilter', 'value'),
     Input('switch-qa-graph', 'value'),
     Input('switches-qa-procstatus', 'value'),
     Input('switches-qa-modality', 'value'),
     Input('radio-qa-pivot', 'value'),
     Input('button-qa-refresh', 'n_clicks')])
def update_qa(
    selected_proc,
    selected_scan,
    selected_sess,
    selected_proj,
    selected_starttime,
    selected_endtime,
    selected_autofilter,
    selected_graph,
    selected_procstatus,
    selected_modality,
    selected_pivot,
    n_clicks
):
    graph_content = []
    refresh = False

    logger.debug('update_all')

    # Load. This data will already be merged scans and assessors, row per
    if utils.was_triggered('button-qa-refresh'):
        # Refresh data if refresh button clicked
        logger.debug('refresh:clicks={}'.format(n_clicks))
        refresh = True

    logger.debug(f'loading data:{selected_proj}')
    df = load_data(
        projects=selected_proj,
        refresh=refresh,
        hidetypes=selected_autofilter)

    # Truncate NOTE
    if 'NOTE' in df:
        df['NOTE'] = df['NOTE'].str.slice(0, 70)

    # Update lists of possible options for dropdowns (could have changed)
    # make these lists before we filter what to display
    proj, sess, proc, scan = load_options(df)

    # Remove from selected what is no longer an option
    if selected_sess:
        selected_sess = [x for x in selected_sess if x in sess]
    if selected_proc:
        selected_proc = [x for x in selected_proc if x in proc]
    if selected_scan:
        selected_scan = [x for x in selected_scan if x in scan]

    # Convert to dash options
    proj = utils.make_options(proj)
    sess = utils.make_options(sess)
    proc = utils.make_options(proc)
    scan = utils.make_options(scan)

    # Filter data based on dropdown values
    df = data.filter_data(
        df,
        selected_proj,
        selected_proc,
        selected_scan,
        selected_starttime,
        selected_endtime,
        selected_sess)

    if not df.empty and selected_modality:
        df = df[df.MODALITY.isin(selected_modality + ['SGP'])]

    if not df.empty and selected_procstatus:
        df = df[df.STATUS.isin(selected_procstatus)]

    if df.empty:
        records = []
        columns = []
    elif selected_pivot == 'proj':
        # Get the qa pivot from the filtered data
        dfp = qa_pivot(df)

        if selected_graph:
            logger.debug('making graph')
            # Graph it
            graph_content = _get_graph_content(dfp)

        # Make the table data
        selected_cols = ['PROJECT']

        dfp = dfp.reset_index()

        if selected_proc:
            selected_cols += selected_proc
            show_proc = [x for x in selected_proc if x in dfp.columns]
        else:
            show_proc = []

        if selected_scan:
            selected_cols += selected_scan
            show_scan = [x for x in selected_scan if x in dfp.columns]
        else:
            show_scan = []

        show_col = show_proc + show_scan

        if show_col:

            # aggregrate to most common value (mode)
            if 'E' in selected_procstatus:
                dfp = dfp.fillna('E')
                dfp = dfp.pivot_table(
                    index=('PROJECT'),
                    values=show_col,
                    aggfunc=lambda x: x.mode().iat[0],
                    fill_value='E',
                )
            else:
                dfp = dfp.fillna('')
                dfp = dfp.pivot_table(
                    index=('PROJECT'),
                    values=show_col,
                    aggfunc=lambda x: x.mode().iat[0],
                    fill_value=np.nan,
                )

            # Replace chars with emojis
            for p in show_col:
                dfp[p] = dfp[p].str.replace('P', '✅')
                dfp[p] = dfp[p].str.replace('X', '🩷')
                dfp[p] = dfp[p].str.replace('Q', '🟩')
                dfp[p] = dfp[p].str.replace('N', '🟡')
                dfp[p] = dfp[p].str.replace('R', '🔷')
                dfp[p] = dfp[p].str.replace('F', '❌')
                dfp[p] = dfp[p].str.replace('E', '□')

            # Drop empty rows
            dfp = dfp.replace('', np.nan)
            dfp = dfp.dropna(subset=show_col)
        else:
            # Exclude SGP
            dfp = dfp[dfp.MODALITY != 'SGP']
            dfp = dfp[dfp.SESSTYPE != 'SGP']

            dfp = dfp.sort_values('MODALITY')

            typecount = len(dfp.SESSTYPE.unique())
            if typecount < 10:
                # Get column names
                selected_cols += ['SESSIONS'] + list(dfp.SESSTYPE.unique())

                dfp2 = dfp[['PROJECT', 'SESSTYPE', 'MODALITY']].copy()
                dfp2 = dfp2.drop_duplicates()
                dfp2['SESSIONS'] = dfp2['MODALITY'].map(MOD2EMO).fillna('?')
                dfp2 = dfp2.pivot_table(
                    index=('PROJECT'),
                    values='SESSIONS',
                    aggfunc=lambda x: ''.join(x))

                dfp = dfp.pivot_table(
                    index=('PROJECT'),
                    columns='SESSTYPE',
                    values='SESSION',
                    aggfunc='count',
                    fill_value='',
                )

                # And smack it together now
                dfp = dfp.merge(dfp2, left_index=True, right_index=True)
            else:
                selected_cols += list(dfp.MODALITY.unique())

                dfp = dfp.pivot_table(
                    index=('PROJECT'),
                    columns='MODALITY',
                    values='SESSION',
                    aggfunc='count',
                    fill_value='',
                )

        # Format as column names and record dictionaries for dash table
        columns = utils.make_columns(selected_cols)
        records = dfp.reset_index().to_dict('records')

    elif selected_pivot == 'subj':
        # row per subject

        # Get the qa pivot from the filtered data
        dfp = qa_pivot(df)

        if selected_graph:
            # Make graphs
            graph_content = _get_graph_content(dfp)

        # Get the table
        dfp = dfp.reset_index()

        selected_cols = ['PROJECT', 'SUBJECT']

        if selected_proc:
            selected_cols += selected_proc
            show_proc = [x for x in selected_proc if x in dfp.columns]
        else:
            show_proc = []

        if selected_scan:
            selected_cols += selected_scan
            show_scan = [x for x in selected_scan if x in dfp.columns]
        else:
            show_scan = []

        show_col = show_proc + show_scan

        if show_col:
            # append sess type to proctype/scantype columns
            # before agg so we get a column per sesstype
            # but only if there are fewer than 10 session types
            typecount = len(dfp.SESSTYPE.unique())
            if typecount < 10:
                # agg to most common value (mode) per sesstype per show_col
                dfp = dfp.fillna('')
                dfp = dfp.pivot_table(
                    index=('PROJECT', 'SUBJECT', 'SUBJECTLINK'),
                    columns='SESSTYPE',
                    values=show_col,
                    aggfunc=lambda x: x.mode().iat[0],
                    fill_value='',
                )

                if typecount > 1:
                    # Concatenate column levels to get one level with delimiter
                    dfp.columns = [f'{c[1]}_{c[0]}' for c in dfp.columns.values]
                else:
                    dfp.columns = [c[0] for c in dfp.columns.values]

                # Drop empty
                for p in dfp.columns:
                    dfp[p] = dfp[p].astype(str).str.replace(
                        '[]', '', regex=False)

                # Really drop empty
                dfp = dfp.replace(r'^\s*$', np.nan, regex=True)

                # For real drop empty
                dfp = dfp.dropna(axis=1, how='all')

                # Save the list columns before we reset index
                show_col = list(dfp.columns)

                # Clear the index so all columns are named
                dfp = dfp.reset_index()
            else:
                # aggregrate to most common value (mode)
                dfp = dfp.pivot_table(
                    index=('PROJECT', 'SUBJECT', 'SUBJECTLINK'),
                    values=show_col,
                    aggfunc=lambda x: x.mode().iat[0],
                )

            for p in show_col:
                dfp[p] = dfp[p].str.replace('P', '✅')
                dfp[p] = dfp[p].str.replace('X', '🩷')
                dfp[p] = dfp[p].str.replace('Q', '🟩')
                dfp[p] = dfp[p].str.replace('N', '🟡')
                dfp[p] = dfp[p].str.replace('R', '🔷')
                dfp[p] = dfp[p].str.replace('F', '❌')
                if 'E' in selected_procstatus:
                    dfp[p] = dfp[p].fillna('□')

            # Drop empty rows
            dfp = dfp.dropna(subset=show_col)

            selected_cols = ['PROJECT', 'SUBJECT'] + show_col
        else:
            # No types selected so show session types by modality

            dfp = dfp[dfp.MODALITY != 'SGP']

            dfp = dfp.sort_values('MODALITY')

            dfp['EMO'] = dfp['MODALITY'].map(MOD2EMO).fillna('?')

            # Pivot to column for each session type
            show_col = list(dfp.SESSTYPE.unique())
            selected_cols = ['SUBJECT', 'PROJECT'] + show_col
            dfp = dfp.pivot_table(
                index=('SUBJECT', 'PROJECT', 'SUBJECTLINK'),
                values='EMO',
                columns='SESSTYPE',
                aggfunc=lambda x: ''.join(x))

            for p in show_col:
                if 'E' in selected_procstatus:
                    dfp[p] = dfp[p].fillna('□')

        # Format as column names and record dictionaries for dash table
        columns = utils.make_columns(selected_cols)
        records = dfp.reset_index().to_dict('records')

        # Format records
        for r in records:
            if r['SUBJECT'] and 'SUBJECTLINK' in r:
                _subj = r['SUBJECT']
                _link = r['SUBJECTLINK']
                r['SUBJECT'] = f'[{_subj}]({_link})'

        # Format columns
        for i, c in enumerate(columns):
            if c['name'] in ['SESSION', 'SUBJECT']:
                columns[i]['type'] = 'text'
                columns[i]['presentation'] = 'markdown'
    else:
        # Get the qa pivot from the filtered data
        dfp = qa_pivot(df)

        if selected_graph:
            graph_content = _get_graph_content(dfp)

        # Get the table data
        selected_cols = [
            'SESSION', 'SUBJECT', 'PROJECT', 'DATE', 'SESSTYPE', 'SITE']

        if selected_proc:
            selected_cols += selected_proc
            show_proc = [x for x in selected_proc if x in dfp.columns]
        else:
            show_proc = []

        if selected_scan:
            selected_cols += selected_scan
            show_scan = [x for x in selected_scan if x in dfp.columns]
        else:
            show_scan = []

        show_col = show_proc + show_scan

        for p in show_col:
            # Replace letters with emojis
            dfp[p] = dfp[p].str.replace('P', '✅')
            dfp[p] = dfp[p].str.replace('X', '🩷')
            dfp[p] = dfp[p].str.replace('Q', '🟩')
            dfp[p] = dfp[p].str.replace('N', '🟡')
            dfp[p] = dfp[p].str.replace('R', '🔷')
            dfp[p] = dfp[p].str.replace('F', '❌')
            if 'E' in selected_procstatus:
                dfp[p] = dfp[p].fillna('□')

        # Drop empty rows
        dfp = dfp.dropna(subset=show_col)

        # Final column is always notes
        selected_cols.append('NOTE')

        # Format as column names and record dictionaries for dash table
        columns = utils.make_columns(selected_cols)
        records = dfp.reset_index().to_dict('records')

        # Format records
        for r in records:
            if r['SESSION'] and 'SESSIONLINK' in r:
                _sess = r['SESSION']
                _link = r['SESSIONLINK']
                r['SESSION'] = f'[{_sess}]({_link})'

            if r['SUBJECT'] and 'SUBJECTLINK' in r:
                _subj = r['SUBJECT']
                _link = r['SUBJECTLINK']
                r['SUBJECT'] = f'[{_subj}]({_link})'

        # Format columns
        for i, c in enumerate(columns):
            if c['name'] in ['SESSION', 'SUBJECT']:
                columns[i]['type'] = 'text'
                columns[i]['presentation'] = 'markdown'

    # Count how many rows are in the table
    _count = len(records)
    if _count > 1:
        rowcount = '{} rows'.format(len(records))
    else:
        rowcount = ''

    # Return table, figure, dropdown options
    logger.debug('update_qa:returning data')

    return [proc, scan, sess, proj, records, columns, graph_content, rowcount, rowcount]



"""Wrapper functions for dcm2nii"""
import logging
import subprocess as sb
import os


def dicom2nifti(dicomdir):
    """ convert dicom to nifti + json using dcm2niix """
    cmd = f'dcm2niix -9 -ba n -z o -f %s_%d {dicomdir}'
    logging.debug(f'running cmd:{cmd}')
    try:
        sb.run(cmd, shell=True)
    except sb.CalledProcessError:
        return []

    niftis = [f for f in os.listdir(dicomdir) if f.endswith('.nii.gz')]

    return niftis


def rename_dicom(in_dir, out_dir):
    # Use "dcm2niix -r" to sort the dicoms into series folders
    # with each file named with instance number zero-padded to 5 digits
    logging.debug('creating renamed/sorted dicom...')
    try:
        sb.run([
            'dcm2niix',
            '-r',
            'y',
            '-f',
            '%s/%5r.dcm',
            '-o',
            out_dir,
            in_dir,
        ])
    except sb.CalledProcessError as err:
        logging.error(f'error:{err}')


def sanitize_filename(filename):
    _dir = os.path.dirname(filename)
    _old = os.path.basename(filename)
    _new = "".join([x if (x.isalnum() or x == '.') else "_" for x in _old])
    if _old != _new:
        # Rename with the sanitized filename
        os.rename(os.path.join(_dir, _old), os.path.join(_dir, _new))
        filename = os.path.join(_dir, _new)

    return filename



import json


SCAN_URI = '/REST/experiments?xsiType=xnat:imagesessiondata\
&columns=\
ID,\
label,\
project,\
subject_label,\
xnat:imagesessiondata/acquisition_site,\
xnat:imagescandata/id,\
xnat:imagescandata/type,\
xnat:imagescandata/quality,\
xnat:imagesessiondata/date'


def process_project(xnat, project, relabels):
    """Apply relabels to project scans"""
    results = relabel_scans(xnat, project, relabels)
    return results


def relabel_scans(xnat, project, relabels):
    results = []

    # get a list of scans from the project
    scan_uri = '{}&project={}'.format(SCAN_URI, project)
    json_data = json.loads(xnat._exec(scan_uri, 'GET'), strict=False)
    scan_list = json_data['ResultSet']['Result']

    # iterate scan and relabel if needed
    for cur_scan in scan_list:
        scan_type = cur_scan['xnat:imagescandata/type']
        if scan_type in relabels:
            proj = cur_scan['project']
            subj = cur_scan['subject_label']
            sess = cur_scan['label']
            scan = cur_scan['xnat:imagescandata/id']

            scan_obj = xnat.select_scan(proj, subj, sess, scan)
            if not scan_obj.exists():
                print('nope')
                continue

            # Set the new type
            new_type = relabels[scan_type]
            scan_obj.attrs.set('xnat:imagescandata/type', new_type)
            results.append({
                'description': f'xnat_relabel_scans:{new_type}',
                'result': 'COMPLETE',
                'category': 'xnat_relabel_scans',
                'subject': subj,
                'session': sess,
                'scan': scan})

    return results



import logging
import os
from datetime import datetime

import pandas as pd

from .. import utils
from ...garjus import Garjus


logger = logging.getLogger('dashboard.activity.data')


# This is where we save our cache of the data
def get_filename():
    datadir = f'{Garjus().cachedir()}/DATA'
    filename = f'{datadir}/activitydata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def get_data(proj_filter):
    df = pd.DataFrame()
    dfc = pd.DataFrame()
    dfq = pd.DataFrame()
    dfj = pd.DataFrame()

    # This week: monday of this week
    # import datetime
    # today = datetime.date.today()
    # startdate = today - datetime.timedelta(days=today.weekday())

    # This month: first date of current month
    # startdate = datetime.datetime.today().replace(day=1)

    # Past month
    from dateutil.relativedelta import relativedelta
    startdate = datetime.today() - relativedelta(months=1)
    startdate = startdate.strftime('%Y-%m-%d')

    g = Garjus()

    if not g.redcap_enabled():
        return None

    logger.info(f'loading activity:startdate={startdate}')
    dfc = g.activity(startdate=startdate)
    dfx = g.assessors()

    dfq = load_recent_qa(dfx, startdate=startdate)
    logger.info('loaded {} qa records'.format(len(dfq)))

    dfj = load_recent_jobs(dfx, startdate=startdate)
    logger.info('loaded {} job records'.format(len(dfj)))

    # Concatentate all the dataframes into one
    df = pd.concat([dfc, dfq, dfj], ignore_index=True)
    df.sort_values(by=['DATETIME'], inplace=True, ascending=False)
    df.reset_index(inplace=True)
    df['ID'] = df.index

    return df


def load_recent_qa(df, startdate):
    df = df.copy()

    df['LABEL'] = df['ASSR']
    df['CATEGORY'] = df['PROCTYPE']

    # Filter by qc date
    df = df[df['QCDATE'] >= startdate]

    df['STATUS'] = df['QCSTATUS'].map({
        'Failed': 'FAIL',
        'Passed': 'PASS'}).fillna('UNKNOWN')

    df['SOURCE'] = 'qa'

    df['CATEGORY'] = df['PROCTYPE']

    df['DESCRIPTION'] = 'QA' + ':' + df['LABEL']

    df['DATETIME'] = df['QCDATE']

    return df


def load_recent_jobs(df, startdate):
    df = df.copy()
    df['LABEL'] = df['ASSR']
    df['CATEGORY'] = df['PROCTYPE']

    # Filter by jobstartdate date, include anything with job running
    df = df[(df['JOBDATE'] >= startdate) | (df['PROCSTATUS'] == 'JOB_RUNNING')]

    df['STATUS'] = df['PROCSTATUS'].map({
        'COMPLETE': 'COMPLETE',
        'JOB_FAILED': 'FAIL',
        'JOB_RUNNING': 'NPUT'}).fillna('UNKNOWN')

    df['SOURCE'] = 'dax'

    df['CATEGORY'] = df['PROCTYPE']

    df['DESCRIPTION'] = 'JOB' + ':' + df['LABEL']

    df['DATETIME'] = df['JOBDATE']

    return df


def run_refresh(filename):
    proj_filter = []

    df = get_data(proj_filter)

    utils.save_data(df, filename)

    return df


def load_data(refresh=False):
    filename = get_filename()

    if refresh or not os.path.exists(filename):
        run_refresh(filename)

    logger.info('reading data from file:{}'.format(filename))
    return utils.read_data(filename)


def filter_data(df, projects, categories, sources):
    # Filter by project
    if projects:
        logger.debug('filtering by project:')
        logger.debug(projects)
        df = df[df['PROJECT'].isin(projects)]

    # Filter by category
    if categories:
        logger.debug('filtering by category:')
        logger.debug(categories)
        df = df[(df['CATEGORY'].isin(categories))]

    # Filter by source
    if sources:
        logger.debug('filtering by source:')
        logger.debug(sources)
        df = df[(df['SOURCE'].isin(sources))]

    return df



import os
import logging
import sys
import io
import re
import datetime
import fnmatch
import tempfile
from datetime import datetime

import pandas as pd
import numpy as np
import redcap


logger = logging.getLogger('garjus.automations.ACOBA_crt')


API_URL = 'https://redcap.vanderbilt.edu/api/'
API_KEY = ''

FORM_INCOMPLETE = 0
FORM_UNVERIFIED = 1
FORM_COMPLETE = 2

REDCAP_EVENTS = ['screening_arm_1', 'baseline_arm_1']

REDCAP_FIELDS = [
    'participant_id',
    'crt_file',
    'crt_recog_mean',
    'crt_recog_median',
    'crt_motor_mean',
    'crt_motor_median',
    'crt_total_mean',
    'crt_total_median',
    'crt_notes',
    'choice_reaction_time_complete',
]


def reformat_date(date_str, in_format, out_format):
    return datetime.strptime(date_str,in_format).strftime(out_format)


def parse_crt(crt_file):
    logger.debug(f'parsing CRT:{crt_file}')
    
    xl = pd.ExcelFile(crt_file)
    _df = xl.parse(xl.sheet_names[0], header=None)
    _data = dict()
           
    if _df.iloc[0][0] != 'Choice Reaction Time Results:':
        logger.error('invalid CRT file')
        return None

    date_str = str(_df.iloc[1][0].split(' ')[1])
    _data['crt_date'] = reformat_date(date_str,'%m/%d/%Y', '%Y-%m-%d')
    _data['crt_num_trials'] = str(int(_df.iloc[2][2]))
    _data['crt_mean_recog_time'] = str(_df.iloc[6][3])
    _data['crt_mean_motor_time'] = str(_df.iloc[7][3])
    _data['crt_mean_tot_time'] = str(_df.iloc[8][3])
    _data['crt_median_recog_time'] = str(_df.iloc[6][5])
    _data['crt_median_motor_time'] = str(_df.iloc[7][5])
    _data['crt_median_tot_time'] = str(_df.iloc[8][5])
    return _data


def download_file(proj, record_id, event_id, field_id, tmpdir):
    filename = os.path.join(tmpdir, record_id+'_'+event_id+'_'+field_id)
    
    logger.debug(f'downloading file:{record_id}:{event_id}:{field_id}')
    
    try:
        cont,hdr = proj.export_file(record=record_id, event=event_id, field=field_id)
        if cont == '':
            raise RedcapError
    except RedcapError:
        logger.error('downloading file')
        return None
    else:
        with open(filename, 'wb') as f:
            f.write(cont)

        return filename


def extract_crt(proj, record_id, event_id, file_field, tmpdir):
    _file = download_file(proj, record_id, event_id, file_field, tmpdir)

    if _file is None:
        return None
   
    _data = parse_crt(_file)
    
    return _data


def load_crt(proj, record_id, event_id, file_field, tmpdir):
    logger.debug('extracting:'+file_field)
    crt_data = extract_crt(proj, record_id, event_id, file_field, tmpdir)
    if crt_data is None:
        logger.error(f'extract failed:{record_id}')
        return

    # Import the data
    data = {}
    data[proj.def_field] = record_id
    data['redcap_event_name'] = event_id
    data['crt_recog_mean'] = crt_data['crt_mean_recog_time']
    data['crt_recog_median'] = crt_data['crt_median_recog_time'] 
    data['crt_motor_mean'] = crt_data['crt_mean_motor_time'] 
    data['crt_motor_median'] = crt_data['crt_median_motor_time'] 
    data['crt_total_mean'] = crt_data['crt_mean_tot_time'] 
    data['crt_total_median'] = crt_data['crt_median_tot_time'] 
    data['choice_reaction_time_complete'] = FORM_COMPLETE
    try:
        response = proj.import_records([data])
        assert 'count' in response
        logger.debug(f'CRT uploaded:{record_id}')
    except AssertionError as err:
        logger.error(f'CRT upload:{record_id}:{err}')



logging.basicConfig(
    format='%(asctime)s - %(levelname)s:%(name)s:%(message)s',
    level=logging.DEBUG,
    datefmt='%Y-%m-%d %H:%M:%S')

tmpdir = tempfile.mkdtemp()
logger.debug(f'tmpdir:{tmpdir}')

# Load the REDCap database
logger.info('INFO:loading REDCap...')
proj = redcap.Project(API_URL, API_KEY)
data = proj.export_records(events=REDCAP_EVENTS, fields=REDCAP_FIELDS)

for r in data:
    if not r['crt_file']:
        logger.debug(f'no file:{r}')
        continue

    if not str(r['crt_file']).endswith('.xlsx'):
        logger.debug(f'wrong format:{r}')
        continue

    if r['crt_recog_mean']:
        logger.debug(f'DEBUG:already extracted:{r[proj.def_field]}:{r["crt_file"]}:{r["crt_recog_mean"]}')
        continue

    logger.info(f'loading:{r["crt_file"]}')
    load_crt(proj, r[proj.def_field], r['redcap_event_name'], 'crt_file', tmpdir)
       
logger.info('DONE!')



"""

Garjus automations.

Automation names correspond to folder name.

"""
import logging
import importlib
import tempfile

from ..utils_redcap import download_file, field2events


# TODO: move slice timing to redcap


logger = logging.getLogger('garjus.automations')


D3_SLICE_TIMING = [
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40,
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40,
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40]

DM2_SLICE_TIMING = [
    0.0, 0.65, 0.065, 0.7150000000000001, 0.13, 0.78, 0.195, 0.845, 0.26, 0.91,
    1.235, 0.585, 1.17, 0.52, 1.105, 0.455, 1.04, 0.39, 0.9750000000000001,
    0.325,
    0.0, 0.65, 0.065, 0.7150000000000001, 0.13, 0.78, 0.195, 0.845, 0.26, 0.91,
    1.235, 0.585, 1.17, 0.52, 1.105, 0.455, 1.04, 0.39, 0.9750000000000001,
    0.325,
    0.0, 0.65, 0.065, 0.7150000000000001, 0.13, 0.78, 0.195, 0.845, 0.26, 0.91,
    1.235, 0.585, 1.17, 0.52, 1.105, 0.455, 1.04, 0.39, 0.9750000000000001,
    0.325]

REMBRANDT_SLICE_TIMING = [
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40,
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40,
    0.00, 0.80, 0.08, 0.88, 0.16, 0.96, 0.24, 1.04, 0.32, 1.12,
    1.52, 0.72, 1.44, 0.64, 1.36, 0.56, 1.28, 0.48, 1.20, 0.40]


def update(garjus, projects, autos_include=None, autos_exclude=None):
    """Update project progress."""
    for p in projects:
        logging.debug(f'updating automations:{p}')
        update_project(garjus, p, autos_include, autos_exclude)


def update_project(garjus, project, autos_include=None, autos_exclude=None):
    """Update automations for project."""
    scan_autos = garjus.scan_automations(project)
    etl_autos = garjus.etl_automations(project)
    edat_autos = garjus.edat_automation_choices()

    if autos_include:
        # Apply include filter
        scan_autos = [x for x in scan_autos if x in autos_include]
        etl_autos = [x for x in etl_autos if x in autos_include]
        edat_autos = [x for x in edat_autos if x in autos_include]

    if autos_exclude:
        # Apply exclude filter
        scan_autos = [x for x in scan_autos if x not in autos_exclude]
        etl_autos = [x for x in etl_autos if x not in autos_exclude]
        edat_autos = [x for x in edat_autos if x not in autos_exclude]

    if scan_autos:
        logging.debug(f'running scan automations:{project}:{scan_autos}')
        _run_scan_automations(scan_autos, garjus, project)

    for a in etl_autos:
        logging.debug(f'running automation:{project}:{a}')
        _run_etl_automation(a, garjus, project)

    _run_edat_automations(edat_autos, garjus, project)


def _run_edat_automations(automations, garjus, project):
    results = []
    edats = garjus.edat_protocols(project)
    scanp = garjus.scanning_protocols(project)
    primary_redcap = garjus.primary(project)
    limbo = garjus.project_setting(project, 'limbodir')
    scans = garjus.scans(projects=[project]).to_dict('records')
    convertdir = garjus.project_setting(project, 'convertdir')
    event2sess = {}

    if primary_redcap is None:
        logger.error(f'primary redcap not found, check keys:{project}')
        return

    for p in scanp:
        s = p['scanning_xnatsuffix']
        _events = [x.strip() for x in p['scanning_events'].split(',')]
        for e in _events:
            event2sess[e] = s

    # load the automations
    try:
        edat_limbo2redcap = importlib.import_module(
            f'garjus.automations.edat_limbo2redcap')
        edat_convert2tab = importlib.import_module(
            f'garjus.automations.edat_convert2tab')
        edat_redcap2xnat = importlib.import_module(
            f'garjus.automations.edat_redcap2xnat')
    except ModuleNotFoundError as err:
        logger.error(f'error loading modules:{err}')
        return

    for e in edats:
        edat_autos = [
            k.split('edat_autos___')[1]
            for k, v in e.items() if k.startswith('edat_autos') and v == '1']

        # Apply filter
        edat_autos = [x for x in edat_autos if x in automations]

        _events = [x.strip() for x in e['edat_events'].split(',')]
        _nums = [x.strip() for x in e['edat_eventnums'].split(',')]
        _event2num = dict(zip(_events, _nums))

        if 'edat_limbo2redcap' in edat_autos:
            results += edat_limbo2redcap.process_project(
                primary_redcap,
                _events,
                e['edat_rawfield'],
                e['edat_convfield'],
                e['edat_fileprefix'],
                limbo,
                event2sess=_event2num)

        if 'edat_convert2tab' in edat_autos and convertdir:
            results += edat_convert2tab.process_project(
                primary_redcap,
                _events,
                e['edat_rawfield'],
                e['edat_convfield'],
                convertdir)

        if 'edat_redcap2xnat' in edat_autos:
            results += edat_redcap2xnat.process_project(
                garjus.xnat(),
                primary_redcap,
                _events,
                e['edat_convfield'],
                event2sess,
                scans,
                e['edat_scantype'],
                'EDAT')

        if 'edat_etl' in edat_autos:
            if e['edat_convfield'] == 'tat_conv':
                # Trait Adjective Task
                try:
                    etl_traitadjtask = importlib.import_module(
                        'garjus.automations.etl_traitadjtask')

                    logger.info(f'{project}:ETL Trait Adjective Task')
                    results += etl_traitadjtask.process_project(primary_redcap)
                except ModuleNotFoundError as err:
                    logger.error(f'error loading modules:{err}')

            elif e['edat_convfield'] == 'tat_recall_conv':
                # Trait Adjective Recall
                try:
                    etl_traitadjrecall = importlib.import_module(
                        'garjus.automations.etl_traitadjrecall')
                    logger.info(f'{project}:ETL Trait Adjective Recall')
                    results += etl_traitadjrecall.process_project(primary_redcap)
                except ModuleNotFoundError as err:
                    logger.error(f'error loading modules:{err}')
                    return

        # Upload results to garjus
        for r in results:
            r.update({'project': project})
            garjus.add_activity(**r)


def _parse_scanmap(scanmap):
    """Parse scan map stored as string into map."""
    # Parse multiline string of delimited key value pairs into dictionary
    scanmap = dict(x.strip().split(':', 1) for x in scanmap.split('\n'))

    # Remove extra whitespace from keys and values
    scanmap = {k.strip(): v.strip() for k, v in scanmap.items()}

    return scanmap


def _run_etl_automation(automation, garjus, project):
    """Load the project primary redcap."""
    results = []

    project_redcap = garjus.primary(project)
    if not project_redcap:
        logger.debug(f'primary redcap not found:{project}')
        return

    if automation == 'etl_nihexaminer':
        results = _run_etl_nihexaminer(project_redcap)
    else:
        # load the automation
        try:
            m = importlib.import_module(f'garjus.automations.{automation}')
        except ModuleNotFoundError as err:
            logger.error(f'error loading module:{automation}:{err}')
            return

        # Run it
        try:
            results = m.process_project(project_redcap)
        except Exception as err:
            logger.error(f'{project}:{automation}:failed to run:{err}')
            return

    # Upload results to garjus
    for r in results:
        r.update({'project': project, 'category': automation})
        r.update({'description': r.get('description', automation)})
        garjus.add_activity(**r)


def _run_etl_nihexaminer(project):
    """Process examiner files from REDCap and upload results."""
    data = {}
    results = []
    events = []
    fields = []
    records = []
    flank_field = 'flanker_file'
    nback_field = 'nback_upload'
    shift_field = 'set_shifting_file'
    cpt_field = 'cpt_upload'
    done_field = 'flanker_score'

    # load the automation
    try:
        examiner = importlib.import_module(f'garjus.automations.etl_nihexaminer')
    except ModuleNotFoundError as err:
        logger.error(f'error loading module:examiner:{err}')
        return

    if 'flanker_summfile' in project.field_names:
        # Alternate file field names
        flank_field = 'flanker_summfile'
        nback_field = 'nback_summfile'
        shift_field = 'set_shifting_summfile'
        cpt_field = 'cpt_summfile'

    # Get the fields
    fields = [
        project.def_field,
        done_field,
        cpt_field,
        nback_field,
        shift_field,
        flank_field,
        'dot_count_tot',
        'anti_trial_1',
        'anti_trial_2',
        'correct_f',
        'correct_l',
        'correct_animal',
        'correct_veg',
        'repetition_f',
        'rule_vio_f',
        'repetition_l',
        'rule_vio_l',
        'repetition_animal',
        'rule_vio_animal',
        'repetition_veg',
        'rule_vio_veg',
        'brs_1',
        'brs_2',
        'brs_3',
        'brs_4',
        'brs_5',
        'brs_6',
        'brs_7',
        'brs_8',
        'brs_9',
    ]

    if 'correct_s' in project.field_names:
        fields.extend([
            'correct_s', 'rule_vio_s', 'repetition_s',
            'correct_t', 'rule_vio_t', 'repetition_t',
            'correct_fruit', 'rule_vio_fruit', 'repetition_fruit',
            'correct_r', 'rule_vio_r', 'repetition_r',
            'correct_m', 'rule_vio_m', 'repetition_m',
            'correct_cloth', 'rule_vio_cloth', 'repetition_cloth',
        ])

    # Determine events
    events = field2events(project, cpt_field)

    # Get records for those events and fields
    records = project.export_records(fields=fields, events=events)

    for r in records:
        data = {}
        record_id = r[project.def_field]
        event_id = r['redcap_event_name']

        if r[done_field]:
            logger.debug(f'already ETL:{record_id}:{event_id}')
            continue

        if not r[cpt_field]:
            logger.debug(f'no data file:{record_id}:{event_id}')
            continue

        # Check for blanks
        has_blank = False
        check_fields = [
            flank_field,
            nback_field,
            shift_field,
            cpt_field,
            'dot_count_tot',
            'anti_trial_1',
            'anti_trial_2',
            'correct_f',
            'correct_l',
            'correct_animal',
            'correct_veg',
            'repetition_f',
            'rule_vio_f',
            'repetition_l',
            'rule_vio_l',
            'repetition_animal',
            'rule_vio_animal',
            'repetition_veg',
            'rule_vio_veg',
            'brs_1',
            'brs_2',
            'brs_3',
            'brs_4',
            'brs_5',
            'brs_6',
            'brs_7',
            'brs_8',
            'brs_9']

        for k in check_fields:
            if r[k] == '' and k != done_field:
                logger.debug(f'blank value:{record_id}:{event_id}:{k}')
                has_blank = True
                break

        if has_blank:
            continue

        logger.debug(f'running nihexaminer ETL:{record_id}:{event_id}')

        # Get values needed for scoring
        manual_values = {
            'dot_total': int(r['dot_count_tot']),
            'anti_trial_1': int(r['anti_trial_1']),
            'anti_trial_2': int(r['anti_trial_2']),
            'cf1_corr': int(r['correct_animal']),
            'cf1_rep': int(r['repetition_animal']),
            'cf1_rv': int(r['rule_vio_animal']),
            'brs_1': int(r['brs_1']),
            'brs_2': int(r['brs_2']),
            'brs_3': int(r['brs_3']),
            'brs_4': int(r['brs_4']),
            'brs_5': int(r['brs_5']),
            'brs_6': int(r['brs_6']),
            'brs_7': int(r['brs_7']),
            'brs_8': int(r['brs_8']),
            'brs_9': int(r['brs_9']),
        }

        if r['correct_f']:
            # examiner version 0
            manual_values.update({
                'vf1_corr': int(r['correct_f']),
                'vf1_rep': int(r['repetition_f']),
                'vf1_rv': int(r['rule_vio_f']),
                'vf2_corr': int(r['correct_l']),
                'vf2_rep': int(r['repetition_l']),
                'vf2_rv': int(r['rule_vio_l']),
                'cf2_corr': int(r['correct_veg']),
                'cf2_rep': int(r['repetition_veg']),
                'cf2_rv': int(r['rule_vio_veg'])
            })
        elif r['correct_t']:
            # examiner version 1
            manual_values.update({
                'vf1_corr': int(r['correct_t']),
                'vf1_rep': int(r['repetition_t']),
                'vf1_rv': int(r['rule_vio_t']),
                'vf2_corr': int(r['correct_s']),
                'vf2_rep': int(r['repetition_s']),
                'vf2_rv': int(r['rule_vio_s']),
                'cf2_corr': int(r['correct_fruit']),
                'cf2_rep': int(r['repetition_fruit']),
                'cf2_rv': int(r['rule_vio_fruit'])
            })
        else:
            # examiner version 2
            manual_values.update({
                'vf1_corr': int(r['correct_r']),
                'vf1_rep': int(r['repetition_r']),
                'vf1_rv': int(r['rule_vio_r']),
                'vf2_corr': int(r['correct_m']),
                'vf2_rep': int(r['repetition_m']),
                'vf2_rv': int(r['rule_vio_m']),
                'cf2_corr': int(r['correct_cloth']),
                'cf2_rep': int(r['repetition_cloth']),
                'cf2_rv': int(r['rule_vio_cloth'])
            })

        with tempfile.TemporaryDirectory() as tmpdir:
            # Get files needed
            flank_file = f'{tmpdir}/flanker.csv'
            cpt_file = f'{tmpdir}/cpt.csv'
            nback_file = f'{tmpdir}/nback.csv'
            shift_file = f'{tmpdir}/shift.csv'

            try:
                # Download files from redcap
                logger.debug(f'download files:{record_id}:{event_id}:{flank_file}')
                download_file(project, record_id, flank_field, flank_file, event_id=event_id)
                logger.debug(f'download NBack:{record_id}:{event_id}:{nback_field}')
                download_file(project, record_id, nback_field, nback_file, event_id=event_id)
                logger.debug(f'download Shift:{record_id}:{event_id}:{shift_field}')
                download_file(project, record_id, shift_field, shift_file, event_id=event_id)
                logger.debug(f'download CPT:{record_id}:{event_id}:{cpt_field}')
                download_file(project, record_id, cpt_field, cpt_file, event_id=event_id)
            except Exception as err:
                logger.error(f'downloading files:{record_id}:{event_id}')
                continue

            try:
                # Process inputs
                data = examiner.process(
                    manual_values,
                    flank_file,
                    cpt_file,
                    nback_file,
                    shift_file)
            except Exception as err:
                logger.error(f'processing examiner:{record_id}:{event_id}:{err}')
                continue

        # Load data back to redcap
        _load(project, record_id, event_id, data)
        results.append({'subject': record_id, 'event': event_id})

    return results


def _run_scan_automations(automations, garjus, project):
    results = []
    proj_scanmap = garjus.project_setting(project, 'scanmap')
    sess_replace = garjus.project_setting(project, 'relabelreplace')
    scan_data = garjus.scanning_protocols(project)
    site_data = garjus.sites(project)
    protocols = garjus.scanning_protocols(project)
    project_redcap = garjus.primary(project)

    # Add slice timing
    if project == 'REMBRANDT':
        logger.debug(f'running add_slicetiming:{project}')

        slicetiming = importlib.import_module(
            'garjus.automations.xnat_add_slicetiming')
        results += slicetiming.process_project(
            garjus,
            project,
            D3_SLICE_TIMING,
            ['fMRI_REST1', 'fMRI_REST2'],
            sites=['VUMC'],
        )
    elif project == 'D3':
        logger.debug(f'running add_slicetiming:{project}')

        slicetiming = importlib.import_module(
            'garjus.automations.xnat_add_slicetiming')
        results += slicetiming.process_project(
            garjus,
            project,
            D3_SLICE_TIMING,
            ['fMRI_REST1', 'fMRI_REST2'],
        )

    # load the automations
    try:
        xnat_auto_archive = importlib.import_module(
            f'garjus.automations.xnat_auto_archive')
        xnat_relabel_sessions = importlib.import_module(
            f'garjus.automations.xnat_relabel_sessions')
        xnat_relabel_scans = importlib.import_module(
            f'garjus.automations.xnat_relabel_scans')
        xnat_dcm2niix = importlib.import_module(
            f'garjus.automations.xnat_dcm2niix')
        xnat_ma3stats2voltxt = importlib.import_module(
            'garjus.automations.xnat_ma3stats2voltxt')
        logger.debug('modules loaded')
    except ModuleNotFoundError as err:
        logger.error(f'error loading scan automations:{err}')
        return

    if 'xnat_auto_archive' in automations and project_redcap and garjus.has_dcm2niix():
        # Apply autos to each scanning protocol
        for p in protocols:
            date_field = p['scanning_datefield']
            sess_field = p['scanning_srcsessfield']
            sess_suffix = p['scanning_xnatsuffix']
            src_project = p['scanning_srcproject']
            alt_primary = p['scanning_altprimary']

            # Get events list
            events = None
            if p.get('scanning_events', False):
                events = [x.strip() for x in p['scanning_events'].split(',')]

            # Make the scan table that links what's entered at the scanner with
            # what we want to label the scans
            if alt_primary:
                scan_redcap = garjus.alternate(alt_primary)
            else:
                scan_redcap = project_redcap

            scan_table = _make_scan_table(
                scan_redcap,
                events,
                date_field,
                sess_field,
                sess_suffix)

            # Run
            logger.debug(f'running xnat_auto_archive:{project}:{events}')
            results += xnat_auto_archive.process_project(
                garjus, scan_table, src_project, project)

    # Apply relabeling
    if 'xnat_relabel_sessions' in automations:
        # Build the session relabling
        sess_relabel = _session_relabels(scan_data, site_data)

        # Run it
        logger.debug(f'{project}:running session relabel')
        results += xnat_relabel_sessions.process_project(
            garjus.xnat(), project, sess_relabel, sess_replace)

    if 'xnat_relabel_scans' in automations and proj_scanmap:
        # Parse scan map
        proj_scanmap = _parse_scanmap(proj_scanmap)

        # Run it
        logger.debug(f'{project}:running scan relabel:{proj_scanmap}')
        results += xnat_relabel_scans.process_project(
            garjus.xnat(), project, proj_scanmap)

    # MA3 stats 2 vol txt
    if 'xnat_ma3stats2voltxt' in automations:
        logger.info(f'running ma3stats2voltxt:{project}')
        assessors = garjus.assessors(
            projects=[project], proctypes=['Multi_Atlas_v3'])
        xnat = garjus.xnat()

        # Add resource information, note takes a minute to load resources
        _resources = garjus.assessor_resources(project, 'Multi_Atlas_v3')
        assessors['RESOURCES'] = assessors.apply(
            lambda x: _resources.get(x.ASSR, ''), axis=1)

        # Run it
        results += xnat_ma3stats2voltxt.process_project(
            xnat,
            project,
            assessors)

    # d2n
    if garjus.has_dcm2niix() and 'dcm2niix' in automations:
        logger.debug(f'{project}:running dcm2niix')
        results += xnat_dcm2niix.process_project(garjus, project)

    # Upload results to garjus
    for r in results:
        r['project'] = project
        garjus.add_activity(**r)


def _make_scan_table(
    project,
    events,
    date_field,
    sess_field,
    scan_suffix,
):
    """Make the scan table, linking source to destination subject/session."""
    data = []
    id2subj = {}

    # Shortcut
    def_field = project.def_field

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get mri records from redcap
    fields = [date_field, sess_field, def_field]
    try:
        rec = project.export_records(fields=fields, events=events)
    except Exception as err:
        logger.error(err)
        return []

    # Only if date is entered
    rec = [x for x in rec if x[date_field]]

    # Only if entered
    rec = [x for x in rec if x[sess_field]]

    # Set the subject and session
    for r in rec:
        d = {}
        try:
            d['dst_subject'] = id2subj.get(r[def_field])
        except KeyError:
            logger.warn(f'blank subject number:{r[def_field]}')
            continue

        if not d['dst_subject']:
            logger.warn(f'blank subject number:{r[def_field]}')
            continue

        d['src_session'] = r[sess_field].strip()
        d['src_subject'] = d['src_session']
        d['dst_session'] = d['dst_subject'] + scan_suffix
        data.append(d)

    return data


def _session_relabels(scan_data, site_data):
    """Build session relabels."""
    relabels = []

    # Build the session relabeling from scan_autos and sites
    for rec in scan_data:
        relabels.append((
            'session_label',
            '*' + rec['scanning_xnatsuffix'],
            'session_type',
            rec['scanning_xnattype']))

    for rec in site_data:
        relabels.append((
            'session_label',
            rec['site_sessmatch'],
            'site',
            rec['site_shortname']))

    return relabels


def _load(project, record_id, event_id, data):
    data[project.def_field] = record_id
    data['redcap_event_name'] = event_id
    data = {k: str(v) for k, v in data.items()}

    try:
        response = project.import_records([data])
        assert 'count' in response
        logger.debug(f'uploaded:{record_id}:{event_id}')
    except AssertionError as e:
        logger.error('error uploading', record_id, e)



import logging


logger = logging.getLogger('garjus.issues.audit_edat')


def audit(
    project,
    events,
    raw_field,
    tab_field,
    ready_field,
):
    results = []
    id2subj = {}
    def_field = project.def_field
    fields = [def_field, raw_field, tab_field, ready_field]

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}

    # Get records
    records = project.export_records(fields=fields, events=events)

    # Process each record
    for rec in records:
        record_id = rec[def_field]
        event = rec['redcap_event_name']
        subj = id2subj.get(record_id, record_id)

        # Skip if not ready
        if not rec[ready_field]:
            logger.debug(f'{subj}:{event}:not ready yet')
            continue

        # already has converted
        if rec[tab_field]:
            logger.debug(f'{subj}:{event}:converted found')
            continue

        # Check for edat file
        if not rec[raw_field]:
            # Missing edat
            logger.debug(f'{subj}:{event}:missing edat')
            results.append({
                'category': 'MISSING_EDAT',
                'subject': subj,
                'event': event,
                'field': raw_field,
                'description': 'need to run edat2tab or manually upload'})
            continue

        # Check for missing data flag
        if 'MISSING_DATA' in rec[raw_field]:
            logger.debug('{}:{}:{}'.format(subj, event, 'missing data flag'))
            continue

        # Check for converted edat file
        if not rec[tab_field]:
            # Missing converted edat
            logger.debug(f'{subj}:{event}:missing converted edat')
            results.append({
                'category': 'MISSING_CONVERTED_EDAT',
                'subject': subj,
                'event': event,
                'field': tab_field,
                'description': 'need to check converter vm'})
            continue

    return results



import logging

from dash import dcc, html, dash_table as dt, Input, Output
import dash_bootstrap_components as dbc

from ..app import app
from .. import utils
from . import data
from ...dictionary import COLUMNS

logger = logging.getLogger('dashboard.analyses')


COMPLETE2EMO = {'0': '🔴', '1': '🟡', '2': '🟢'}


def get_content():
    #columns = utils.make_columns(COLUMNS.get('analyses'))
    columns = utils.make_columns([
        'PROJECT',
        'ID',
        'NAME',
        'COMPLETE',
        'EDIT',
        'INPUT',
        'OUTPUT',
    ])

    # Format columns with links as markdown text
    for i, c in enumerate(columns):
        if c['name'] in ['OUTPUT', 'EDIT', 'INPUT']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    content = [
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-analyses-proj',
                    multi=True,
                    placeholder='Select Project(s)',
                ),
                width=3,
            ),
        ),
        dbc.Spinner(id="loading-analyses-table", children=[
            dbc.Label('Loading...', id='label-analyses-rowcount1'),
        ]),
        dt.DataTable(
            columns=columns,
            data=[],
            page_action='none',
            sort_action='none',
            id='datatable-analyses',
            style_cell={
                'textAlign': 'center',
                'padding': '15px 5px 15px 5px',
                'height': 'auto',
            },
            style_header={
                'fontWeight': 'bold',
            },
            style_cell_conditional=[
                {'if': {'column_id': 'NAME'}, 'textAlign': 'left'},
            ],
            # Aligns the markdown cells, both vertical and horizontal
            css=[dict(selector="p", rule="margin: 0; text-align: center")],
        ),
        html.Label('0', id='label-analyses-rowcount2')]

    return content


def load_analyses(projects=[]):

    if projects is None:
        projects = []

    return data.load_data(projects, refresh=True)


@app.callback(
    [
     Output('dropdown-analyses-proj', 'options'),
     Output('datatable-analyses', 'data'),
     Output('label-analyses-rowcount1', 'children'),
     Output('label-analyses-rowcount2', 'children'),
    ],
    [
     Input('dropdown-analyses-proj', 'value'),
    ])
def update_analyses(
    selected_proj,
):
    logger.debug('update_all')

    # Load selected data with refresh if requested
    df = load_analyses(selected_proj)

    # Get options based on selected projects, only show proc for those projects
    proj_options = data.load_options()

    logger.debug(f'loaded options:{proj_options}')

    proj = utils.make_options(proj_options)

    # Filter data based on dropdown values
    df = data.filter_data(df)

    df['COMPLETE'] = df['COMPLETE'].map(COMPLETE2EMO).fillna('?')

    # Get the table data as one row per assessor
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
        # Make edit a link
        _link = r['EDIT']
        _text = 'edit'
        r['EDIT'] = f'[{_text}]({_link})'

        # Make edit a link
        if not r['INPUT']:
            pass
        else:
            _link = r['INPUT']
            _text = r['INPUT'].rsplit('/', 1)[1]
            r['INPUT'] = f'[{_text}]({_link})'

        # Make output a link
        if not r['OUTPUT']:
            pass
        elif 'sharepoint.com' in r['OUTPUT']:
            _link = r['OUTPUT']
            _text = 'OneDrive'
            r['OUTPUT'] = f'[{_text}]({_link})'
        elif 'xnat' in r['OUTPUT']:
            _link = r['OUTPUT']
            _text = r['OUTPUT'].rsplit('/', 1)[1]
            r['OUTPUT'] = f'[{_text}]({_link})'
        else:
            r['OUTPUT'] = r['OUTPUT']

    # Count how many rows are in the table
    rowcount = '{} rows'.format(len(records))

    return [proj, records, rowcount, rowcount]



"""dax queue 2 garjus queue."""
import subprocess
from io import StringIO
import logging
import os
from datetime import datetime

import pandas as pd

logger = logging.getLogger('dax2garjus')


# TODO: check status jobs showing as RUNNING but not in queue,
# check residr for completed/vs failed, and set to Finished-COMPLETE, 
# Finished-JOB_FAILED or something maybe Finished in blue vs. red?

# TODO: complete job information from slurm, for now we just 
# want to know about open jobs

# This is a temporary bridge between garjus and dax.

USER = 'vuiis_daily_singularity'
SQUEUE = 'squeue -u vuiis_daily_singularity --format="%j|%A|%L|%m|%M|%p|%t|%u|%S|%T|%V|%l|"'
RESDIR = '/nobackup/vuiis_daily_singularity/Spider_Upload_Dir'

JOB_TAB_COLS = [
    'LABEL', 'PROJECT', 'STATUS', 'PROCTYPE', 'USER',
    'JOBID', 'TIME', 'WALLTIME', 'LASTMOD']

SQUEUE_COLS = [
    'NAME', 'ST', 'STATE', 'PRIORITY', 'JOBID', 'MIN_MEMORY',
    'TIME', 'SUBMIT_TIME', 'START_TIME', 'TIME_LIMIT', 'TIME_LEFT', 'USER']

# we concat diskq status and squeue status to make a single status
# squeue states: CG, F, PR, S, ST
# diskq statuses: JOB_RUNNING, JOB_FAILED, NEED_TO_RUN, COMPLETE,
# UPLOADING, READY_TO_COMPLETE, READY_TO_UPLOAD
STATUS_MAP = {
    'COMPLETENONE': 'COMPLETE',
    'JOB_FAILEDNONE': 'FAILED',
    'JOB_RUNNINGCD': 'RUNNING',
    'JOB_RUNNINGCG': 'RUNNING',
    'JOB_RUNNINGF': 'RUNNING',
    'JOB_RUNNINGR': 'RUNNING',
    'JOB_RUNNINGNONE': 'RUNNING',
    'JOB_RUNNINGPD': 'PENDING',
    'NONENONE': 'WAITING',
    'READY_TO_COMPLETENONE': 'COMPLETE',
    'READY_TO_UPLOADNONE': 'COMPLETE'}


def _load_dax_queue():
    logger.debug('loading diskq')
    diskq_df = _load_diskq_queue()

    logger.debug('loading squeue')
    squeue_df = _load_slurm_queue()

    # merge squeue data into task queue
    logger.debug('merging data')

    if diskq_df.empty and squeue_df.empty:
        logger.debug('both empty')
        df = pd.DataFrame(columns=diskq_df.columns.union(squeue_df.columns))
    elif diskq_df.empty:
        logger.debug('diskq empty')
        df = squeue_df.reindex(squeue_df.columns.union(diskq_df.columns), axis=1)
    elif squeue_df.empty:
        logger.debug('squeue empty')
        df = diskq_df.reindex(diskq_df.columns.union(squeue_df.columns), axis=1)
    else:
        df = pd.merge(diskq_df, squeue_df, how='outer', on=['LABEL', 'USER'])

    if not df.empty:
        # assessor label is delimited by "-x-", first element is project,
        # fourth element is processing type
        df['PROJECT'] = df['LABEL'].str.split('-x-', n=1, expand=True)[0]
        df['PROCTYPE'] = df['LABEL'].str.split('-x-', n=4, expand=True)[3]

        # Add some text to avoid blanks in the table
        df['JOBID'].fillna('not in queue', inplace=True)

        # create a concatenated status that maps to full status
        df['psST'] = df['procstatus'].fillna('NONE') + df['ST'].fillna('NONE')
        df['STATUS'] = df['psST'].map(STATUS_MAP).fillna('UNKNOWN')

    # Determine how long ago status changed
    # how long has it been running, pending, waiting or complete?

    # Minimize columns
    logger.debug('finishing data')
    df = df.reindex(columns=JOB_TAB_COLS)

    return df.sort_values('LABEL')

    return df


# Loads the dax queue from disk
def _load_diskq_queue(status=None):
    task_list = list()
    diskq_dir = os.path.join(RESDIR, 'DISKQ')
    batch_dir = os.path.join(diskq_dir, 'BATCH')

    for t in os.listdir(batch_dir):
        assr = os.path.splitext(t)[0]
        logger.debug(f'load task:{assr}')
        task = _load_diskq_task(diskq_dir, assr)
        task['USER'] = USER
        task_list.append(task)

    if len(task_list) > 0:
        df = pd.DataFrame(task_list)
    else:
        df = pd.DataFrame(columns=[
            'LABEL', 'procstatus', 'jobid', 'jobnode', 'jobstartdate',
            'memused', 'walltimeused', 'WALLTIME', 'LASTMOD', 'USER'])

    return df


# Load a single task/job information from disk
def _load_diskq_task(diskq, assr):
    return {
        'LABEL': assr,
        'procstatus': _get_diskq_attr(diskq, assr, 'procstatus'),
        'jobid': _get_diskq_attr(diskq, assr, 'jobid'),
        'jobnode': _get_diskq_attr(diskq, assr, 'jobnode'),
        'jobstartdate': _get_diskq_attr(diskq, assr, 'jobstartdate'),
        'memused': _get_diskq_attr(diskq, assr, 'memused'),
        'walltimeused': _get_diskq_attr(diskq, assr, 'walltimeused'),
        'WALLTIME': _get_diskq_walltime(diskq, assr),
        'LASTMOD': _get_diskq_lastmod(diskq, assr)}


# Load slurm data
def _load_slurm_queue():
    try:
        cmd = SQUEUE
        result = subprocess.run([cmd], shell=True, stdout=subprocess.PIPE)
        _data = result.stdout.decode('utf-8')
        df = pd.read_csv(
            StringIO(_data), delimiter='|', usecols=SQUEUE_COLS)
        df['LABEL'] = df['NAME'].str.split('.slurm').str[0]
        return df
    except pd.errors.EmptyDataError:
        return pd.DataFrame(columns=SQUEUE_COLS+['LABEL'])


def _get_diskq_walltime(diskq, assr):
    COOKIE = "#SBATCH --time="
    walltime = None
    bpath = os.path.join(diskq, 'BATCH', assr + '.slurm')

    try:
        with open(bpath, 'r') as f:
            for line in f:
                if line.startswith(COOKIE):
                    walltime = line.split('=')[1].replace('"', '').replace("'", '')
                    break
    except IOError:
        logger.warn('file does not exist:' + bpath)
        return None
    except PermissionError:
        logger.warn('permission error reading file:' + bpath)
        return None

    return walltime


def _get_diskq_lastmod(diskq, assr):

    if os.path.exists(os.path.join(diskq, 'procstatus', assr)):
        apath = os.path.join(diskq, 'procstatus', assr)
    elif os.path.exists(os.path.join(diskq, 'BATCH', assr + '.slurm')):
        apath = os.path.join(diskq, 'BATCH', assr + '.slurm')
    else:
        return None

    updatetime = datetime.fromtimestamp(os.path.getmtime(apath))
    delta = datetime.now() - updatetime
    return delta


def _get_diskq_attr(diskq, assr, attr):
    apath = os.path.join(diskq, attr, assr)

    if not os.path.exists(apath):
        return None

    try:
        with open(apath, 'r') as f:
            return f.read().strip()
    except PermissionError:
        return None


def _get_changes(garjus_queue, dax_queue):
    # Make list of (ID,PROJECT,STATUS) where status doesn't match
    df = pd.merge(garjus_queue, dax_queue, left_on='ASSESSOR', right_on='LABEL')
    df = df[df.STATUS_x != df.STATUS_y]
    df = df[['ID', 'PROJECT_x', 'STATUS_y']]
    df = df.rename(columns={'PROJECT_x': 'PROJECT', 'STATUS_y': 'STATUS'})

    return df


def _get_xnat_changes(garjus_queue, assessors):
    # Make list of (ID,PROJECT,STATUS) where status doesn't match
    df = pd.merge(garjus_queue, assessors, left_on='ASSESSOR', right_on='ASSR')

    # Only where status mismatch
    df = df[df.STATUS != df.PROCSTATUS]

    # Subset of columns we need
    df = df[['ID', 'PROJECT_x', 'PROCSTATUS']]

    # Don't revert to JOB_RUNNING
    df = df[df.PROCSTATUS != 'JOB_RUNNING']

    # Use column names expected by garjus
    df = df.rename(columns={'PROJECT_x': 'PROJECT', 'PROCSTATUS': 'STATUS'})

    return df


def dax2queue(garjus):
    # load diskq, run squeue to get updates, compare to g queue, apply changes
    # in a single call

    resdir = RESDIR

    if not os.path.isdir(resdir):
        raise FileNotFoundError(f'upload directory not found:{resdir}')

    # Get the garjus queue as stored in redcap
    gqueue = garjus.tasks()

    # Get the dax queue on disk
    dqueue = _load_dax_queue()

    # Filter projects to only those garjus knows
    dqueue = dqueue[dqueue.PROJECT.isin(garjus.projects())]

    # Get the changes to apply
    df1 = _get_changes(gqueue, dqueue)

    # Get updates from XNAT (if no longer in dax queue), complete or failed
    df2 = _get_xnat_changes(gqueue, garjus.assessors())

    # Combine dataframes
    df = pd.concat([df1, df2])

    # Apply changes
    if df.empty:
        logger.info('no changes to apply')
    else:
        logger.info('applying status changes')
        garjus.set_task_statuses(df)



"""NIH Examiner data extraction."""
import logging
import tempfile
import subprocess
import os

import pandas as pd


# CPT Summary File
cpt_columns = [
    'target_corr',
    'target_errors',
    'nontarget_corr',
    'nontarget_errors',
    'target_mean',
    'target_median',
    'target_stdev',
    'performance_errors']


# Flanker Summary File
flank_columns = [
    'congr_corr',
    'congr_mean',
    'congr_median',
    'congr_stdev',
    'incongr_corr',
    'incongr_mean',
    'incongr_median',
    'incongr_stdev',
    'total_corr',
    'total_mean',
    'total_median',
    'total_stdev',
    'flanker_score',
    'flanker_error_diff']


# N-back Summary File
nback_columns = [
    'nb1_score',
    'nb1_bias',
    'nb1_corr',
    'nb1_errors',
    'nb1_mean',
    'nb1_median',
    'nb1_stdev',
    'nb2_score',
    'nb2_bias',
    'nb2_corr',
    'nb2_errors',
    'nb2_mean',
    'nb2_median',
    'nb2_stdev',
    'nb2int_corr',
    'nb2int_errors',
    'nb2int_mean',
    'nb2int_median',
    'nb2int_stdev']


# Output columns we want to keep
scoring_columns = [
    'executive_composite',
    'executive_se',
    'fluency_factor',
    'fluency_se',
    'cog_control_factor',
    'cog_control_se',
    'working_memory_factor',
    'working_memory_se']


# Columns required by scoring program
input_columns = [
    'subject_id',
    'session_date',
    'site_id',
    'session_num',
    'language',
    'age',
    'dot_total',
    'nb1_score',
    'nb2_score',
    'flanker_score',
    'error_score',
    'antisacc',
    'shift_score',
    'vf1_corr',
    'vf2_corr',
    'cf1_corr',
    'cf2_corr']


def process(
    manual_values,
    flank_file,
    cpt_file,
    nback_file,
):
    """Process NIH Examiner files and return subset of data."""
    mv = manual_values

    # Extract data from files
    flank_data = _extract_onerow_file(flank_file)
    cpt_data = _extract_onerow_file(cpt_file)
    nback_data = _extract_onerow_file(nback_file)

    antisacc = mv['anti_trial_1'] + mv['anti_trial_2']

    behav_total = 0
    try:
        for i in range(9):
            b = int(mv[f'brs_{i+1}'])
            if b < 4:
                behav_total += b
    except Exception as err:
        logging.error(err)
        return

    # Calculate error score
    error_score = \
        cpt_data['nontarget_errors'] + \
        flank_data['flanker_error_diff'] + \
        mv['vf1_rep'] + \
        mv['vf1_rv'] + \
        mv['vf2_rep'] + \
        mv['vf2_rv'] + \
        mv['cf1_rep'] + \
        mv['cf1_rv'] + \
        mv['cf2_rep'] + \
        mv['cf2_rv'] + \
        behav_total

    # Collect inputs
    inputs = {
        'subject_id': '',
        'session_date': '',
        'site_id': '',
        'session_num': '',
        'language': '1',
        'age': '',
        'dot_total': mv['dot_total'],
        'antisacc': antisacc,
        'vf1_corr': mv['vf1_corr'],
        'vf2_corr': mv['vf2_corr'],
        'cf1_corr': mv['cf1_corr'],
        'cf2_corr': mv['cf2_corr'],
        'nb1_score': nback_data['nb1_score'],
        'nb2_score': nback_data['nb2_score'],
        'flanker_score': flank_data['flanker_score'],
        'error_score': error_score,
        'shift_score': '',
    }

    # Run the Examiner Scoring Program
    with tempfile.TemporaryDirectory() as tmpdir:
        print(inputs)
        score_data = _scoring(inputs, tmpdir)
        print(score_data)

    # Transform data for upload
    return _transform(flank_data, cpt_data, nback_data, score_data)


def _transform(flank_data, cpt_data, nback_data, score_data):
    """Take data extracted from files and prep for REDCap"""
    data = {}
    data.update(_subset(flank_data, flank_columns))
    data.update(_subset(cpt_data, cpt_columns))
    data.update(_subset(nback_data, nback_columns))
    data.update(_subset(score_data, scoring_columns))
    return data


def _extract_onerow_file(filename):
    """Extract data from file that has a header row and one data row"""
    try:
        df = pd.read_csv(filename)
    except:
        df = pd.read_excel(filename)

    if len(df) > 1:
        logging.warn('multiple rows, using last!')

    # Get data from last row
    return df.iloc[-1].to_dict()


def _subset(data, columns):
    """Return subset of key/values based on column or key names specified."""
    return {k: v for k, v in data.items() if k in columns}


def _scoring(inputs, tmpdir):
    """Run the Examiner Scoring program and return outputs"""
    inputfile = f'{tmpdir}/input.csv'
    outputfile = f'{tmpdir}/output.csv'

    input_data = [str(inputs.get(x, '')) for x in input_columns]

    # Save input file
    with open(inputfile, 'w') as f:
        f.write(','.join(input_columns))
        f.write('\n')
        f.write(','.join(input_data))
        f.write('\n')
        f.write(','.join(input_data))
        f.write('\n')

    # Run the scoring program with the input file
    _wd = os.path.expanduser('~/git/garjus/src/automations/etl_nihexaminer/Scoring')
    _script = 'examiner_scoring.R'
    cmd = f'Rscript --vanilla'
    cmd += f' -e "setwd(\'{_wd}\')"'
    cmd += f' -e "source(\'{_script}\')"'
    cmd += f' -e "score_file(\'{inputfile}\', \'{outputfile}\')"'
    res = subprocess.call(cmd, shell=True)

    # Load values from output file
    return _extract_onerow_file(outputfile)


if __name__ == "__main__":
    logging.basicConfig( level=logging.DEBUG)

    _dir = '7500_wk12'
    flank_file = f'{_dir}/Flanker_Summary_7500_2_12_23_2021_13h_15m.csv'
    cpt_file = f'{_dir}/CPT_Summary_7500_2_12_23_2021_13h_34m.csv'
    nback_file = f'{_dir}/NBack_Summary_7500_2_12_23_2021_13h_40m.csv'

    # Get values needed for scoring
    mv = {
        'dot_total': 17,
        'anti_trial_1': 20,
        'anti_trial_2': 20,
        'cf1_corr': 20,
        'cf1_rep': 1,
        'cf1_rv': 0,
        'brs_1': 4,
        'brs_2': 4,
        'brs_3': 4,
        'brs_4': 4,
        'brs_5': 4,
        'brs_6': 4,
        'brs_7': 4,
        'brs_8': 4,
        'brs_9': 4,
        'vf1_corr': 17,
        'vf1_rep': 0,
        'vf1_rv': 4,
        'vf2_corr': 16,
        'vf2_rep': 1,
        'vf2_rv': 0,
        'cf2_corr': 14,
        'cf2_rep': 2,
        'cf2_rv': 0,
    }

    data = process(mv, flank_file, cpt_file, nback_file)
    import pprint
    pprint.pprint(data)
    pprint.pprint(data.keys())
